/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*
 * getinfo.c
 * Copyright (c) 1999 Madge Networks Ltd.
 *
 * TEG 14/09/1999
 */
/* Please refer to the LICENSE file supplied with this source code */

#include <stdio.h>
#include <linux/pci.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <errno.h>
#include <netdb.h>

#include <unistd.h>
#include <malloc.h>
#include <string.h>

#include "madgepci.h"
#include "mdgioctl.h"
#include "bia.h"

#define PCI_PREFIX	"/proc/bus/pci"
#define PCI_FILE 	"devices"

/* There is 256 bytes of config data but only the first 64 are
   accessible to non root users */
#define PCI_CONFIG_LEN	(256)

pci_device *head = NULL;

void madge_pci_append(pci_device *dev)
{
	pci_device * curdev;
	if (head == NULL)
	{
		head = dev;
		dev->next = dev->prev = NULL;
		return;
	}
	curdev = head;
	while(curdev->next) curdev=curdev->next;
	curdev->next = dev;
	dev->prev = curdev;
	dev->next = NULL;
	return;
}

pci_device * madge_pci_get_list()
{
	return head;
}

void madge_pci_insert(pci_device *dev)
{
	pci_device * curdev = head;

	while(curdev && curdev->info.busdevfn < dev->info.busdevfn)
		curdev=curdev->next;

	if (curdev)
	{
		/* Need to insert in order */
		if (curdev->prev == NULL)
		{
			dev->next = head;
			dev->prev = NULL;
			head = dev;
		}
		else
		{
			dev->next = curdev;
			curdev->prev->next = dev;
			dev->prev = curdev->prev;
			curdev->prev = dev;
		}
	}
	else
	{
		/* Got to end of list just append to end */
		madge_pci_append(dev);
	}
}

pci_device * madge_pci_new_device(int busdevfn)
{
	pci_device * dev;
	if (NULL == (dev = madge_pci_find_busdevfn(busdevfn)))
	{
		if (NULL == (dev = malloc(sizeof(pci_device))))
		{
			fprintf(stderr,"No memory\n");
			return NULL;
		}
		memset(dev,0,sizeof(pci_device));
		dev->info.busdevfn = busdevfn;
		madge_pci_insert(dev);
	}
	return dev;
}


pci_device * madge_pci_find_busdevfn(int busdevfn)
{
	pci_device * curdev = head;
	while(curdev)
	{
		if (curdev->info.busdevfn == busdevfn)
			return curdev;
		curdev=curdev->next;
	}
	return NULL;
}


int madge_check_for_mtok_module()
{
	FILE *fptr;
		
	if (NULL == (fptr = fopen("/proc/modules","r")))
	{
		return 0;
	}	                                                                
	while(!feof(fptr))
	{
		char Buffer[50];
		fgets(Buffer,50,fptr);
		if (strncmp(Buffer,"mtok",4) == 0)
		{
			fclose(fptr);
			return 1;
		}
	}
	fclose(fptr);
	return 0;
}

int madge_check_for_proc_mtok()
{
	struct stat buf;
	if (stat("/proc/net/mtok",&buf) < 0)
 		return 0;
 	else
 		return 1;
}

int madge_pci_get_old_running_info()
{
	FILE * fptr;
	int Ver1,Ver2,Ver3;
	int Count = 0;
	char Name[50],laabuf[14],biabuf[14];
	
	if (NULL == (fptr = fopen("/proc/net/mtok","rb")))
	{
		return 0;	
	}

	fscanf(fptr,"%d.%d.%d\n",&Ver1,&Ver2,&Ver3);
	
	while(!feof(fptr))
	{
		int retval,type,subtype,iobase,xfermode,irq,ring_speed,openmode,busdevfn;
		pci_device * dev;
		 
		retval = fscanf(fptr,"%s %d %d %s %s %X %X %d %d %d %d %*X\n",	
				Name,			/* Device Name (e.g. tr0) 		 */
				&type,			/* Adapter Type 				 */
				&subtype,		/* Adapter Subtype 				 */
				biabuf,			/* BIA 							 */
				laabuf,			/* LAA 							 */
				&iobase,		/* IO Base 						 */
				&busdevfn,		/* PCI Number (if high bit set)  */
				&xfermode,		/* Transfer Mode (UDT constant)  */
				&irq,			/* IRQ 							 */
				&ring_speed,	/* Ring speed (UDT Constant)     */
				&openmode		/* Open Mode (e.g. DTR, Classic) */
				);

		if (retval != 11 || retval == EOF)
			break;

		busdevfn &= 0x00FFFFFF;
		busdevfn = (busdevfn & 0xFF00) | ((busdevfn & 0xFF) << 3);

		dev = madge_pci_new_device(busdevfn);

		dev->cardid = mtok_pci_type_to_cardid(type,subtype);
		dev->sw_running = 1;
		dev->sw_version = Ver1 << 16 | Ver2 << 8 | Ver3;
		dev->info.iobase = iobase;
		dev->info.irq = irq;
		dev->info.ringspeed = 	(ring_speed == 3 /*RS_4M*/ )   ? 4 :
								(ring_speed == 4 /*RS_16M*/ )  ? 16 :
								(ring_speed == 5 /*RS_100M*/ ) ? 100 : 0;

		printf("ringspeed=%d\n",dev->info.ringspeed);

		dev->info.dtr = (openmode == 2 /*DTR_ONLY*/) ? 1 : 
						(openmode == 1 /*CLASSIC_ONLY*/) ? 2 : 0 ;
						
		madge_parse_bia(biabuf,dev->info.bia);
		madge_parse_bia(laabuf,dev->info.laa);
	}

	fclose(fptr);
	return Count;
}

int madge_pci_get_adapters_v22()
{
	char pcifname[128];
	
	FILE * fptr;

	sprintf(pcifname,"%s/%s",PCI_PREFIX,PCI_FILE);

	if ((fptr = fopen(pcifname,"r")) == NULL)
	{
		return 1;
	}
	
	while(!feof(fptr))
	{
		pci_device * dev;
		char fname[256];
		char pci_config[256];
		FILE *fconfig;
		int manfdev;
		int busdevfn;
		int irq;
		int baseaddr;
		int ret;
		int vendorid;
		int deviceid;
		int ssysid;
		int rev;
		
		mtok_card_id * cardid;
		
		if ((ret = fscanf(fptr,"%X %X %X %X %*X %*X %*X %*X %*X %*X\n",&busdevfn,&manfdev,&irq,&baseaddr)) < 0)
		{
			perror("fscanf");
			return 1;
		}

		vendorid = (manfdev & 0xFFFF0000) >> 16;
		deviceid = (manfdev & 0x0000FFFF);

		if (vendorid != PCI_VENDOR_ID_MADGE)
			continue;

		sprintf(fname,"%s/%02x/%02x.%01x",PCI_PREFIX,PCI_BUS(busdevfn),PCI_SLOT(busdevfn),PCI_FUNC(busdevfn));

		if ((fconfig = fopen(fname,"rb")) == NULL)
		{
			return 1;
		}
		
		if ((ret = fread(pci_config,PCI_CONFIG_LEN,1,fconfig)) < 0)
		{
			perror("Read PCI config space");
			return 1;
		}
		
		if (ret != 1)
		{
			return 1;
		}

		fclose(fconfig);
		
		/* 2.0.x incorrectly defines PCI_SUBSYSTEM_ID so don't use it here */
		ssysid 	= *((unsigned short int *)&pci_config[0x2e]);
		rev		= *((unsigned char *)&pci_config[PCI_REVISION_ID]);
		
		if (NULL == (cardid = mtok_pci_to_cardid(vendorid,deviceid,0,ssysid,rev)))
		{
			return 1;
		}

		if (NULL == (dev = madge_pci_new_device(busdevfn)))
		{
			return 1;
		}
		
		dev->info.busdevfn = busdevfn;
		dev->info.irq  = irq;
//		dev->devid = deviceid;
		dev->info.iobase = baseaddr & PCI_BASE_ADDRESS_IO_MASK;
		dev->cardid = cardid;
		
		/* BIA is stored at offset 0xec on KNOSSOS adapters */
		if (dev->cardid->adapter_type == ADAPTER_TYPE_KNOSSOS)
			memcpy(dev->info.bia,&pci_config[0xec],6);
	}
	fclose(fptr);	
	return 0;
}


/*
 * Returns:-
 *      0 - Success
 *      1 - Fail but carry on looking
 *      2 - Fail and stop searching other TR devices
 */
int madge_pci_get_one_new_running_info(char *device)
{
	int s;
	struct ifreq ifr;
	MTOK_IOCTL_STATS stats;
	MTOK_IOCTL_INFO	info;
	MTOK_IOCTL_MDGID mdgid;
	pci_device * dev;
	
	memset(&stats,0,sizeof(MTOK_IOCTL_STATS));
	memset(&info,0,sizeof(MTOK_IOCTL_INFO));

	if ((s = socket(AF_INET,SOCK_DGRAM,0)) < 0)
	{
		perror("Error creating socket");
	    return 0;
	}
	
	/* Firstly, establish if this device is controlled
	   by a Madge driver and which version number if so
	 */

	strcpy(ifr.ifr_name,device);
	ifr.ifr_data = (void *)&mdgid;
	mdgid.len = sizeof(MTOK_IOCTL_MDGID);
	mdgid.sig = MTOK_ID_IN;

	if (ioctl(s,MTOK_SIOCMDGID,&ifr) < 0)
	{
		if (errno == ENODEV)
		{
			/* Error no device - reached the end of the line */
			return 2;
		}
		if (errno == EINVAL)
		{
			/* We just queried a TR device which doesn't support private IOCTLs */
			return 1;
		}
		/* A proper error */
		perror("MTOK_SIOCMDGID");
		return 2;
	}
	
	if (mdgid.sig != MTOK_ID_OUT)
	{
		/* Not a Madge Device - continue looping there may be others */
		return 1;
	}
	
	/*
	 * Now query the driver for statistics and information
	 */
 
	ifr.ifr_data = (void *)&info;
	info.len = sizeof(MTOK_IOCTL_INFO);
	info.sig = MTOK_ID_IN;

	if (ioctl(s,MTOK_SIOCADAPTERINFO,&ifr)<0)
	{
		/* Some sort of error in the querying the device - 
		 * still need to carry on looking so return 1
		 */
		perror("MTOK_SIOCINFO");
		return 1;
	}                            

	ifr.ifr_data = (void *)&stats;
	stats.len = sizeof(MTOK_IOCTL_STATS);
	stats.sig = MTOK_ID_IN;

	if (ioctl(s,MTOK_SIOCADAPTERSTATS,&ifr)<0)
	{
		/* Some sort of error in the querying the device - 
		 * still need to carry on looking so return 1
		 */
		perror("MTOK_SIOCSTATS");
		return 1;
	}                            

	dev = madge_pci_new_device(info.busdevfn);
	memcpy(&dev->info,&info,sizeof(MTOK_IOCTL_INFO));
	memcpy(&dev->stats,&stats,sizeof(MTOK_IOCTL_STATS));	

	dev->info.ringspeed = 	(dev->info.ringspeed == 3 /*RS_4M*/ )   ? 4 :
								(dev->info.ringspeed == 4 /*RS_16M*/ )  ? 16 :
								(dev->info.ringspeed == 5 /*RS_100M*/ ) ? 100 : 0;

	dev->cardid = mtok_pci_type_to_cardid(info.type,info.subtype);
	dev->sw_running = 1;
	dev->sw_version = mdgid.version;
	close(s);
	return 0;
}


int madge_pci_get_new_running_info()
{
	int rval;
	char device[8];
	int i = 0;
		do
		{
			sprintf(device,"tr%d",i);
			rval = madge_pci_get_one_new_running_info(device);
			i++;
		} while(rval == 0 || rval==1);
	return 0;
}

