/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*
*
* madgecs.c
*
* Madge Token Ring Device Driver PCMCIA Helper for Linux 2.0.29+.
*
* Copyright (c) Madge Networks Ltd 1998
*
* PVCS Version History:
*
*  $Archive:   W:/PVCSARCH/linux/mtok/pcmcia/madgecs.c_v  $
*  $Revision: 1.5 $
*  $Date: 1999/04/08 14:22:22 $
*  $Author: tom $
*  $Modtime:   05 Jun 1998 13:07:50  $
*    
*****************************************************************************/
/* Please refer to the LICENSE file supplied with this source code */

#include <linux/autoconf.h>

#include <linux/config.h>
#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif

#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/netdevice.h>
#include <linux/malloc.h>

#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>

#include "madgecs.h"
#ifdef DEBUG 
#include "mpoke.h"
#endif 

/* Initialise, add us into list, display details */
extern LINUXDEV * mtok_probe1(LINUXDEV *dev,
                            ADAPTER_ID_T  * padapter_id,int irq,
                            int maxframesize,int txslots,int rxslots,
                            int ringspeed,int dtr,void *usrdata);

/*extern void mtok_cleanup(MTOK_ADAPTER * adap);*/
extern void mtok_add_dev(LINUXDEV *dev);
extern void mtok_remove_dev(LINUXDEV *dev);
extern void mtok_isr IRQ_ARGS(int irq,void * dev_id,struct pt_regs * regs);

void cleanup_module (void);

struct oli_pcmcia {
    RS_TUPLE_T *  olicom_rs_tuple;
    BIA_TUPLE_T * olicom_bia_tuple;
};




/*----------------------------------------------------------------------------
|
| PRODUCT TRACKING STRINGS - These MUST NOT be changed.
|
|---------------------------------------------------------------------------*/

char mtokcs_mver_title[] =
    "VeRsIoN=" MTOKCS_PRODUCT_NAME; 

char mtokcs_mver_copyright[] = 
    "VeRsIoN=" MTOKCS_PRODUCT_COPYRIGHT;

char mtokcs_mver_platform[] = 
    "VeRsIoN=PLATFORM: " MTOKCS_PRODUCT_PLATFORM;

char mtokcs_mver_version[61] =
    "VeRsIoN=VERSION: v" MTOKCS_VERSION;

char mtok_cs_crcp[] = "MaDgE=XXXX";

/*----------------------------------------------------------------------------
|
| Our device information structure.
|
|---------------------------------------------------------------------------*/

dev_info_t mtokcs_dev_info = MTOKCS_NAME;


/*----------------------------------------------------------------------------
|
| Our list of instances.
|
|---------------------------------------------------------------------------*/

MTOKCS_INSTANCE * mtokcs_inst_list = NULL;


/*----------------------------------------------------------------------------
|
| Parameters that can be set with 'insmod'.
|
|---------------------------------------------------------------------------*/

/*
 * Bit map of interrupts to choose from. This means pick from 15, 14, 12, 11, 
 * 10, 9, 7, 5, 4, and 3.
 */
 
static 
u_long irq_mask = 0xdeb8;


/*----------------------------------------------------------------------------
|
| Some pre-declarations.
|
|---------------------------------------------------------------------------*/

dev_link_t * 
mtokcs_attach(void);

void 
mtokcs_detach(dev_link_t * link);

void 
mtokcs_config(dev_link_t * link);

void 
mtokcs_release(u_long arg);

int 
mtokcs_event(event_t event, int priority, event_callback_args_t * args);


/*----------------------------------------------------------------------------
|
| Function    - mtokcs_cs_error
|
| Parameters  - func -> Function number.
|               retc -> Error code.
|
| Purpose     - Inform card services that an error has ocurred.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void 
mtokcs_cs_error(
    int func, 
    int retc
    )
{
    CardServices(ReportError, mtokcs_dev_info, (void *) func, (void *) retc);
}


/*----------------------------------------------------------------------------
|
| Function    - mtokcs_find_registers
|
| Parameters  - inst -> A pointer to our instance structure.
|
| Purpose     - Parse our CIS to find our configuration registers.
|
| Returns     - CS_SUCCESS on success or an error code.
|
|---------------------------------------------------------------------------*/

int
mtokcs_find_registers(
    MTOKCS_INSTANCE * inst
    )
{
    dev_link_t      * link;
    client_handle_t   handle;
    tuple_t           tuple;
    cisparse_t        parse;
    u_char            buf[64];
    int               retc;
    
    link   = &inst->link;
    handle = link->handle;

#ifdef DEBUG
    QPOKE(PO_GREEN,"[mtokcs_find_registers");
#endif

    memset(&tuple,0,sizeof(tuple_t));
    
    /*
     * Read the card's CONFIG tuple to find its configuration registers.
     */
     
    tuple.DesiredTuple = CISTPL_CONFIG;
    
    if ((retc = CardServices(GetFirstTuple, handle, &tuple)) != CS_SUCCESS)
    {
#ifdef DEBUG	
        QPOKE(PO_YELLOW,"GetFirstTuple");
        QPOKEW(PO_RED,"Ret:", retc);
#endif
        mtokcs_cs_error(GetFirstTuple, retc);
        link->state &= ~DEV_CONFIG_PENDING;
#ifdef DEBUG	
        QPOKE(PO_GREEN,"]");
#endif
        return retc;
    }
    
    tuple.TupleData    = buf;
    tuple.TupleDataMax = 64;
    tuple.TupleOffset  = 0;
        
    if ((retc = CardServices(GetTupleData, handle, &tuple)) != CS_SUCCESS)
    {
#ifdef DEBUG	
        QPOKE(PO_YELLOW,"GetTupleData");
        QPOKEW(PO_RED,"Ret:", retc);
#endif
        mtokcs_cs_error(GetTupleData, retc);
        link->state &= ~DEV_CONFIG_PENDING;
#ifdef DEBUG	
        QPOKE(PO_GREEN,"]");
#endif
        return retc;
    }
    
    if ((retc = CardServices(ParseTuple, handle, &tuple, &parse)) != CS_SUCCESS)
    {
#ifdef DEBUG	
        QPOKE(PO_YELLOW,"ParseTuple");
        QPOKEW(PO_RED,"Ret:", retc);
#endif
        mtokcs_cs_error(ParseTuple, retc);
        link->state &= ~DEV_CONFIG_PENDING;
#ifdef DEBUG	 
        QPOKE(PO_GREEN,"]");
#endif
        return retc;
    }
    
    /*
     * Record the configuration register base in the device linkage structure.
     */
     
    link->conf.ConfigBase = parse.config.base;
#ifdef DEBUG	
    QPOKED(PO_CYAN,"ConfigBase:",parse.config.base);
    QPOKE(PO_GREEN,"]");
#endif
    return CS_SUCCESS;
}
    

/*----------------------------------------------------------------------------
|
| Function    - mtokcs_alloc_resources
|
| Parameters  - inst -> A pointer to our instance structure.
|
| Purpose     - Allocate hardware resources for our adapter.
|
| Returns     - CS_SUCCESS on success or an error code.
|
|---------------------------------------------------------------------------*/

int
mtokcs_alloc_resources(
    MTOKCS_INSTANCE * inst
    )
{
    dev_link_t      * link;
    client_handle_t   handle;
    int               retc;
        
    link   = &inst->link;
    handle = link->handle;
        
    /*
     * Try allocating IO ports.
     */
#ifdef DEBUG      
    QPOKE(PO_GREEN,"{RequestIO");
#endif
    link->io.BasePort1 = 0; /* Request any IO base address */
    if ((retc = CardServices(RequestIO, handle, &link->io)) != CS_SUCCESS)
    {
#ifdef DEBUG 
        QPOKEW(PO_RED,"FAIL:",retc);
        QPOKE(PO_GREEN,"}");
#endif
        mtokcs_cs_error(RequestIO,retc);
        return retc;
    }

#ifdef DEBUG 
    QPOKE(PO_GREEN,"}");    

    QPOKED(PO_BLUE,"IO:",  (unsigned) link->io.BasePort1);
#endif
    
    inst->io_init = TRUE;
    
    /*
     * Now allocate an interrupt line.  Note that this does not actually 
     * assign a handler to the interrupt.
     */

#ifdef DEBUG      
    QPOKE(PO_GREEN,"{RequestIRQ");
#endif
    if ((retc = CardServices(RequestIRQ, handle, &link->irq)) != CS_SUCCESS)
    {
#ifdef DEBUG 
        QPOKEW(PO_RED,"Ret:", retc);
#endif
        mtokcs_cs_error(RequestIRQ, retc);
        return retc;
    }        
#ifdef DEBUG 
    QPOKE(PO_GREEN,"}");
    QPOKED(PO_YELLOW,"IRQ:", link->irq.AssignedIRQ);
#endif

    inst->irq_init = TRUE;
    
    /*
     * This actually configures the PCMCIA socket -- setting up the I/O 
     * windows and the interrupt mapping.
     */
     
    if ((retc = CardServices(RequestConfiguration, handle, &link->conf)) != CS_SUCCESS)
    {
#ifdef DEBUG 
        QPOKEW(PO_RED,"Ret:", retc);
#endif
        mtokcs_cs_error(RequestConfiguration, retc);
        return retc;
    }

    inst->sock_init = TRUE;
    
    return CS_SUCCESS;
}

            
/*----------------------------------------------------------------------------
|
| Function    - mtokcs_attach
|
| Parameters  - None.
|
| Purpose     - This function is called by Card Services to create an
|               "instance" of the adapter. It registers with Card
|               services but doesn't actually attempt to initialise an
|               adapter. This happens when we get a card insertion event.
|
| Returns     - A pointer to a dev_link_t structure.
|
|---------------------------------------------------------------------------*/

dev_link_t * 
mtokcs_attach(
    void
    )
{
    MTOKCS_INSTANCE * inst;
    dev_link_t      * link;
    client_reg_t      client_reg;
    int               retc;

#ifdef DEBUG 
    QPOKE(PO_WHITE,"[Attach");
#endif
    
    /* 
     * Allocate an instance control structure.
     */
     
    inst = (MTOKCS_INSTANCE *) kmalloc(sizeof(MTOKCS_INSTANCE), GFP_KERNEL);
    
    if (inst == NULL)
    {
        printk(KERN_ERR MTOKCS_NAME ": Insufficient kernel memory\n");
#ifdef DEBUG 
        QPOKE(PO_WHITE,"]");
#endif
        return NULL;
    }
        
    memset((void *) inst, 0, sizeof(MTOKCS_INSTANCE));

    /*
     * Initialise our device linkage structure.
     */
     
    link       = &inst->link;
    link->priv = (void *) inst;

    /*
     * Setup the release() function.
     */
     
    link->release.function = &mtokcs_release;
    link->release.data     = (u_long) link;

    /* 
     * The io structure describes I/O port mapping.
     */

    /* Madge and Olicom cards require the same values here */
    link->io.NumPorts1   = MTOKCS_IO_RANGE;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
    link->io.IOAddrLines = 16;

    /* 
     * Interrupt setup.
     */

    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1   = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
    link->irq.IRQInfo2   = irq_mask;
	link->irq.Handler    = mtok_isr;
    
    /* 
     * General socket configuration.
     */
     
    link->conf.Attributes  = CONF_ENABLE_IRQ;
    link->conf.Vcc         = 50;
    link->conf.IntType     = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present     = PRESENT_OPTION;

    /*
     * Link this adapter instance into the list of all Madge PCMCIA adapter
     * instances.
     */
     
    inst->next       = mtokcs_inst_list;
    mtokcs_inst_list = inst;
    
    /* 
     * Register with Card Services.
     */
     
    client_reg.dev_info      = &mtokcs_dev_info;
    client_reg.Attributes    = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask     = CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
                               CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
                               CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = mtokcs_event;
    client_reg.Version       = 0x0210;
    
    client_reg.event_callback_args.client_data = link;
    
#ifdef DEBUG 
    QPOKE(PO_GREEN,"{RegisterClient");
#endif
    retc = CardServices(RegisterClient, &link->handle, &client_reg);
#ifdef DEBUG 
    QPOKE(PO_GREEN,"}");
#endif

    if (retc != EOK) 
    {
#ifdef DEBUG 
        QPOKEW(PO_RED,"Ret:",retc);
#endif
        mtokcs_cs_error(RegisterClient, retc);
        mtokcs_detach(link);
#ifdef DEBUG 
        QPOKE(PO_WHITE,"]");
#endif
        return NULL;
    }

    inst->cs_init = TRUE;
#ifdef DEBUG 
    QPOKE(PO_WHITE,"]"); 
#endif
    return link;
} 


/*----------------------------------------------------------------------------
|
| Function    - mtokcs_detach
|
| Parameters  - link -> A pointer to our device linkage structure.
|
| Purpose     - This function is called by Card Services to delete an
|               "instance" of the adapter. If it has been released, all 
|               local data structures are freed. Otherwise, the structures 
|               will be freed when the device is released. 
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void 
mtokcs_detach(
    dev_link_t * link
    )
{
    MTOKCS_INSTANCE   * inst;
    MTOKCS_INSTANCE * * posp;

#ifdef DEBUG 
    QPOKE(PO_WHITE,"[Detach");
#endif

    /*
     * Find our instance structure in the global instance list.
     */

    inst = NULL;
         
    for (posp = &mtokcs_inst_list; *posp != NULL; posp = &((*posp)->next))
    {
        if (*posp == (MTOKCS_INSTANCE *) link->priv)
        {
            inst = (MTOKCS_INSTANCE *) link->priv;
            break;
        }
    }
    
    if (inst == NULL)
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"inst==NULL");
        QPOKE(PO_WHITE,"]"); 
#endif
        return;
    }                
               
    /*
     * If the device is currently configured and active, we won't
     * actually delete it yet.  Instead, it is marked so that when
     * the release() function is called, that will trigger a proper
     * detach().
     */
     
    if ((link->state & DEV_CONFIG) != 0) 
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"Configured");
#endif
        link->state |= DEV_STALE_LINK;
#ifdef DEBUG 
        QPOKE(PO_WHITE,"]");
#endif
        return;
    }

    /* 
     * Break the link with Card Services.
     */
     
    if (inst->cs_init)
    {
        CardServices(DeregisterClient, link->handle);
        inst->cs_init = FALSE;
    }    
    
    /* 
     * Unlink our instance structure from the global list.
     */
    
    *posp = inst->next;
    
    /*
     * Free the instance structure.
     */

    kfree(inst); 
#ifdef DEBUG 
    QPOKE(PO_WHITE,"]");
#endif
}

static int get_tuple(int fn, client_handle_t handle, tuple_t *tuple,
             cisparse_t *parse)
{
    int i;

#ifdef DEBUG 
    QPOKE(PO_CYAN,"[");
#endif
    i = CardServices(fn, handle, tuple);

    if (i != CS_SUCCESS)
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"FAILED");
        QPOKEW(PO_YELLOW,"GetTuple:",fn);
        QPOKEW(PO_RED,"Ret:",i);
        QPOKE(PO_CYAN,"]");
#endif
        return CS_NO_MORE_ITEMS;
    }

    i = CardServices(GetTupleData, handle, tuple);

    if (i != CS_SUCCESS)
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"FAILED");
        QPOKE(PO_YELLOW,"GetTupleData");
        QPOKEW(PO_RED,"Ret:",i);
        QPOKE(PO_CYAN,"]");
#endif
        return i;
    }
    
    if (tuple->DesiredTuple < CISTPL_ORG)
        i = CardServices(ParseTuple, handle, tuple, parse);

#ifdef DEBUG 
    QPOKE(PO_CYAN,"]");
#endif
    return i;
}

#define first_tuple(a, b, c) get_tuple(GetFirstTuple, a, b, c)
#define next_tuple(a, b, c) get_tuple(GetNextTuple, a, b, c)
                                     

/*----------------------------------------------------------------------------
|
| Function    - mtokcs_config
|
| Parameters  - link -> A pointer to our device linkage structure.
|
| Purpose     - This function is scheduled to run after a CARD_INSERTION 
|               event is received, to configure the PCMCIA socket, and to 
|               make the TRN device available to the system.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void 
mtokcs_config(
    dev_link_t * link
    )
{
    MTOKCS_INSTANCE * inst = (MTOKCS_INSTANCE *) link->priv;
    struct oli_pcmcia odata; 
    ADAPTER_ID_T id;
    u_short buf[128];
    u_short manfid,prodid;
    tuple_t tuple;
    client_handle_t handle;
    cisparse_t parse;
            	    
#ifdef DEBUG 
    QPOKE(PO_YELLOW,"[config");
#endif

    handle = link->handle;

    /*
     * Read the card's CONFIG tuple to find its configuration registers.
     */
     
    if (mtokcs_find_registers(inst) != CS_SUCCESS)
    {
        mtokcs_release((u_long) link);
#ifdef DEBUG 
        QPOKE(PO_YELLOW,"]");
#endif
        return;
    }    
    
    /* 
     * Note that we're configuring the adapter.
     */
     
    link->state |= DEV_CONFIG;

    tuple.TupleDataMax = 128;
    tuple.TupleOffset  = 0;
    tuple.DesiredTuple = CISTPL_MANFID;
    tuple.TupleData = (cisdata_t *)buf;
    tuple.Attributes = 0;
    
#ifdef DEBUG 
    QPOKE(PO_BLUE,"{CISTPL_MANFID");
#endif
    if (first_tuple(handle,&tuple,&parse)!=CS_SUCCESS)
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"Fail");
        QPOKE(PO_BLUE,"}");
#endif
        mtokcs_release((u_long) link);
#ifdef DEBUG 
        QPOKE(PO_YELLOW,"]");
#endif
        return;
    }
#ifdef DEBUG 
    QPOKE(PO_BLUE,"}");
#endif

    manfid = le16_to_cpu(buf[0]);
    prodid = le16_to_cpu(buf[1]);

#ifdef DEBUG 
    QPOKEW(PO_YELLOW,"ManfID:",manfid);
    QPOKEW(PO_CYAN,"ProdID:",prodid);
#endif


    switch(manfid)
    {
#ifdef OLICOM_PCMCIA_UDI    
        case PCMCIA_OLICOM_VENDOR_ID:
            /*
             * For Olicom adapters we need to read the BIA and RS
             * information from the CIS.
             */
            tuple.DesiredTuple = RS_TUPLE;
            tuple.TupleData = (cisdata_t *)&inst->olicom_rs_tuple;
            tuple.TupleDataMax = sizeof(RS_TUPLE_T);
            tuple.TupleOffset  = 0;

#ifdef DEBUG 
            QPOKE(PO_BLUE,"{RS_TUPLE");
#endif
            if ((retc=first_tuple(handle,&tuple,&parse))!=CS_SUCCESS)
            {
#ifdef DEBUG 
                QPOKE(PO_RED,"Fail");
                QPOKEW(PO_YELLOW,"Retc:",retc);
                QPOKE(PO_BLUE,"}");
#endif
                mtokcs_release((u_long) link);
#ifdef DEBUG 
                QPOKE(PO_YELLOW,"]");
#endif
                return;
            }
#ifdef DEBUG 
            QPOKE(PO_BLUE,"}");
#endif

            tuple.DesiredTuple = BIA_TUPLE;
            tuple.TupleData = (cisdata_t *)&inst->olicom_bia_tuple;
            tuple.TupleDataMax = sizeof(BIA_TUPLE_T);
            tuple.TupleOffset  = 0;

#ifdef DEBUG 
            QPOKE(PO_BLUE,"{BIA_TUPLE");
#endif
            if ((retc=first_tuple(handle,&tuple,&parse))!=CS_SUCCESS)
            {
#ifdef DEBUG 
                QPOKE(PO_RED,"Fail");
                QPOKEW(PO_YELLOW,"Retc:",retc);
                QPOKE(PO_BLUE,"}");
#endif
                mtokcs_release((u_long) link);
#ifdef DEBUG 
                QPOKE(PO_YELLOW,"]");
#endif
                return;
            }
#ifdef DEBUG 
            QPOKE(PO_BLUE,"}");
#endif

        	id.adapter_type = ADAPTER_TYPE_OLICOM_PCMCIA;

        	switch(prodid)
        	{
        	    case PCMCIA_OLICOM_3221_DEVICE_ID:
                	id.adapter_subtype = ADAPTER_SUBTYPE_OLICOM_PCMCIA_3221;
                break;
                default:
                	/* 
                	 * Other Olicom PCMCIA adapter model numbers 
                	 * 3232, 3230 are not supported yet.
                	 */
#ifdef DEBUG 
                    QPOKE(PO_RED,"!UnknownOlicomDevice!");
#endif
                    mtokcs_release((u_long) link);
#ifdef DEBUG 
                    QPOKE(PO_YELLOW,"]");
#endif
                    return;
            }

        	/* Other Olicom PCMCIA adapter model numbers 3232, 3230 */
            
            /* 
             * Store pointers to the RS and BIA tuples in odata structure.
             * odata is passed to mtok_probe1.  The driver stores the
             * pointers so that in the SSI tuple function the relevant
             * values can be returned.
             */

            odata.olicom_rs_tuple = &inst->olicom_rs_tuple;
            odata.olicom_bia_tuple = &inst->olicom_bia_tuple;

            break;
#endif
        case PCMCIA_MADGE_VENDOR_ID:
        	id.adapter_type = ADAPTER_TYPE_PCMCIA;
            switch(prodid)
            {
                case PCMCIA_MADGE_REV_1_DEVICE_ID:
                	id.adapter_subtype = ADAPTER_SUBTYPE_PCMCIA;
                	break;
                case PCMCIA_MADGE_REV_2_DEVICE_ID:
                	id.adapter_subtype = ADAPTER_SUBTYPE_PCMCIA_2;
                    break;
                default:
#ifdef DEBUG 
                    QPOKE(PO_RED,"!UnknownMadgeDevice!");
#endif
                    mtokcs_release((u_long) link);
#ifdef DEBUG 
                    QPOKE(PO_YELLOW,"]");
#endif
                    return;
            }
            break;

        default:
#ifdef DEBUG 
            QPOKE(PO_RED,"!!UnsupportedAdapter!!");
#endif
            mtokcs_release((u_long) link);
#ifdef DEBUG 
            QPOKE(PO_YELLOW,"]");
#endif
            return;
    }

    /*
     * Allocate hardware resources for the adapter.
     */
     
    if (mtokcs_alloc_resources(inst) != CS_SUCCESS)
    {
        mtokcs_release((u_long) link);
#ifdef DEBUG 
        QPOKE(PO_YELLOW,"]");
#endif
        return;
    }    

    /*
     * Common setup for both Madge and Olicom Cards.
     */

   	id.io_base = link->io.BasePort1;
   	id.adapter_state = ADAPTER_STATE_RESET;
   	id.field_used_to_identify_card = USE_IO_BASE;
   	id.pci_number = 0;

    
    /*
     * Call the driver to register the device with the kernel and initialise 
     * the adapter.
     */
	
	if ( (inst->dev = mtok_probe1(inst->dev,&id,
			link->irq.AssignedIRQ,0,0,0,0,0,&odata)) == NULL )
	
	{
        mtokcs_release((u_long) link);
#ifdef DEBUG 
        QPOKE(PO_YELLOW,"]");
#endif
        return;
	}

    inst->adap_init = TRUE;
     
    /*
     * Initialise the node structure.
     */
     
    link->dev = &inst->node;
       
    /*
     * Note that configuration is no-longer pending.
     */
         
    link->state &= ~DEV_CONFIG_PENDING;
    
#ifdef DEBUG 
    QPOKE(PO_YELLOW,"]");
#endif
}


/*----------------------------------------------------------------------------
|
| Function    - mtokcs_release
|
| Parameters  - arg -> A pointer to our device linkage structure.
|
| Purpose     - After a card is removed, skeleton_release() will unregister 
|               the net device, and release the PCMCIA configuration. If the 
|               device is still open, this will be postponed until it is 
|               closed.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void 
mtokcs_release(
    u_long arg
    )
{
    dev_link_t      * link = (dev_link_t *) arg;
    MTOKCS_INSTANCE * inst = (MTOKCS_INSTANCE *) link->priv;

#ifdef DEBUG 
    QPOKE(PO_GREEN,"[release");
#endif

    if (!link)
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"NULLlink");
        QPOKE(PO_GREEN,"]");
#endif
        return;
    }

#ifdef DEBUG 
    if (link->dev && link->dev->dev_name)
    {
        QPOKE(PO_YELLOW,link->dev->dev_name);
    }
#endif

    /*
     * If the device is currently in use, we won't release until it
     * is actually closed.
     */
     
    if (link->open) 
    {
#ifdef DEBUG 
        QPOKE(PO_RED,"postponed");
        QPOKE(PO_GREEN,"]");
#endif
        link->state |= DEV_STALE_CONFIG;
        return;
    }

    /*
     * Call the driver to closedown the adapter and deregister the device.
     */
     
    if (inst->adap_init)
    { 
	mtok_cleanup(inst->dev->priv);
        inst->adap_init = FALSE;
    }    

    /*
     * Tell the driver the adapter has gone away. It's safe to call this
     * function even if the corresponding mtok_pcmcia_add() has not
     * occurred.
     */
     
    if (inst->dev)
        mtok_remove_dev(inst->dev);
     
    /* 
     * Unlink the device chain.
     */
     
    link->dev = NULL;

    /*
     * Release our resources back to Card Services.
     */
     
    if (inst->sock_init)
    {  
        CardServices(ReleaseConfiguration, link->handle);
        inst->sock_init = FALSE;
    }
    
    if (inst->irq_init)
    {    
        CardServices(ReleaseIRQ, link->handle, &link->irq);
        inst->irq_init = FALSE;
    }
    
    if (inst->io_init)
    {
        CardServices(ReleaseIO, link->handle, &link->io);
        inst->io_init = FALSE;
    }    
    
    /*
     * Note that the instance is not configured.
     */
     
    link->state &= ~DEV_CONFIG;
    
    /*
     * If we've previously postponed a detach() then do the detach() now.
     */
     
    if ((link->state & DEV_STALE_LINK) != 0)
    {
        mtokcs_detach(link);
    }
        
#ifdef DEBUG 
    QPOKE(PO_GREEN,"]");
#endif
}


/*----------------------------------------------------------------------------
|
| Function    - mtokcs_event
|
| Parameters  - event    -> An event code.
|               priority -> The even't priority.
|               args     -> Pointer to an argument structure.
|
| Purpose     - Call to process an event relating to an instance.
|
| Returns     - 0.
|
|---------------------------------------------------------------------------*/

int 
mtokcs_event(
    event_t                 event, 
    int                     priority,
    event_callback_args_t * args
    )
{
    dev_link_t      * link = (dev_link_t *) args->client_data;
    MTOKCS_INSTANCE * inst = (MTOKCS_INSTANCE *) link->priv;
    
#ifdef DEBUG 
    QPOKE(PO_CYAN,"[Event");
#endif
    
    switch (event) 
    {
        /*
         * The adapter has been removed.
         */
         
        case CS_EVENT_CARD_REMOVAL:

#ifdef DEBUG 
            QPOKE(PO_GREEN,"CS_EVENT_CARD_REMOVAL");
#endif
                   
            link->state &= ~DEV_PRESENT;
           
            /*
             * If the instance has been configured then we mark it as busy
             * an schedule a release in a 1/20 of a second (to allow hardware
             * to settle).
             */
             
            if ((link->state & DEV_CONFIG) != 0) 
            {
		MTOK_SET_TBUSY(inst->dev);
		MTOK_RESET_START(inst->dev);
                
                link->release.expires = RUN_AT(HZ/20);
                add_timer(&link->release);
            }
            break;
            
        /*
         * The adapter has been inserted.
         */
             
        case CS_EVENT_CARD_INSERTION:
        
#ifdef DEBUG 
            QPOKE(PO_GREEN,"CS_EVENT_CARD_INSERTION");
            
#endif
            link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
            mtokcs_config(link);
            break;
 
        /*
         * The machine is going into suspend mode.
         */
                    
        case CS_EVENT_PM_SUSPEND:

#ifdef DEBUG 
            QPOKE(PO_GREEN,"CS_EVENT_PM_SUSPEND");
#endif
        
            link->state |= DEV_SUSPEND;
            
            /*
             * If the instance has been configured then kill the adapter
             * and release the configuration.
             */
             
            if ((link->state & DEV_CONFIG) != 0)
            {
		mtok_cleanup(inst->dev->priv);
                inst->adap_init = FALSE;
                
                CardServices(ReleaseConfiguration, link->handle);
            }    
            break;
        
        /*
         * The machine has left suspend mode.
         */
             
        case CS_EVENT_PM_RESUME:
   
#ifdef DEBUG 
            QPOKE(PO_GREEN,"CS_EVENT_PM_RESUME");
#endif
            
            link->state &= ~DEV_SUSPEND;
         
            /*
             * If the instance has been configured then request the 
             * configuration and initialise the adapter.
             */
             
            if ((link->state & DEV_CONFIG) != 0)
            {
                CardServices(RequestConfiguration, link->handle, &link->conf);
                
				mtokcs_config(link);
            }    
            break;
            
        /*
         * The adapter is about to be physically reset.
         */
         
        case CS_EVENT_RESET_PHYSICAL:

#ifdef DEBUG 
            QPOKE(PO_GREEN,"CS_EVENT_RESET_PHYSICAL");
        
#endif
            /*
             * If the instance has been configured then kill the adapter
             * and release the configuration.
             */
             
            if ((link->state & DEV_CONFIG) != 0)
            {
		mtok_cleanup(inst->dev->priv);
                inst->adap_init = FALSE;
                
                CardServices(ReleaseConfiguration, link->handle);
            }    
            break;
            
        /*
         * The adapter has been physically reset.
         */
        
        case CS_EVENT_CARD_RESET:

#ifdef DEBUG 
            QPOKE(PO_GREEN,"CS_EVENT_CARD_RESET");
#endif
        
            /*
             * If the instance has been configured then request the 
             * configuration and initialise the adapter.
             */
             
            if ((link->state & DEV_CONFIG) != 0)
            {
                CardServices(RequestConfiguration, link->handle, &link->conf);
                
				mtokcs_config(link);
            }    
            break;
        
    }
#ifdef DEBUG 
    QPOKE(PO_CYAN,"]");
#endif
    return 0;
}



#if LINUX_VERSION_CODE > 0x20115
MODULE_AUTHOR("Madge Networks (www.madge.com)");
MODULE_DESCRIPTION("Madge Token Ring Adapter driver");
MODULE_PARM(force16,  "i");
#endif

/*---------------------------------------------------------------------------
|
| Function    - init_module
|
| Parameters  - None.
|
| Purpose     - Load time entry point for a dynamically loadable module.
|
| Returns     - 0 on success or an errno.h error code.
|
|--------------------------------------------------------------------------*/

int 
init_module(
    void
    )
{
    servinfo_t serv;

    printk(KERN_NOTICE 
        "Madge Networks Token-Ring Driver PCMCIA Helper v" MTOKCS_VERSION
        " (" __TIME__ " " __DATE__ ")\n");
            
#ifdef MTOKCS_CUSTOMISED_BY
    printk(KERN_NOTICE "Customised by: " MTOKCS_CUSTOMISED_BY "\n");
#endif      
    
    /*
     * Check the Card Services version number.
     */
     
    CardServices(GetCardServicesInfo, &serv);
    
    if (serv.Revision != CS_RELEASE_CODE)
    {
        printk(KERN_WARNING MTOKCS_NAME 
            ": ERROR Card Services release does not match "
            "(is %x should be %x)\n", 
            (unsigned) serv.Revision, 
            (unsigned) CS_RELEASE_CODE);
    }
    
    /*
     * Register with Card Services.
     */
     
    register_pcmcia_driver(&mtokcs_dev_info, &mtokcs_attach, &mtokcs_detach);
    return 0;
}

/*---------------------------------------------------------------------------
|
| Function    - cleanup_module
|
| Parameters  - None.
|
| Purpose     - Unload time entry point for a dynamically loadable module.
|
| Returns     - Nothing.
|
|--------------------------------------------------------------------------*/

void 
cleanup_module(
    void
    )
{
    /*
     * De-register with Card Services.
     */
     
    unregister_pcmcia_driver(&mtokcs_dev_info);
    
    /*
     * Kill of all of the instances.
     */
     
    while (mtokcs_inst_list != NULL) 
    {
        if ((mtokcs_inst_list->link.state & DEV_CONFIG) != 0)
        {
            mtokcs_release((u_long) &mtokcs_inst_list->link);
        }
            
        mtokcs_detach(&mtokcs_inst_list->link);
    }
}


/**** End of madgecs.c ******************************************************/
