/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
*
* madgecs.h
*
* Madge Token Ring Device Driver PCMCIA Helper for Linux 2.0.29+.
*
* Main header file.
*
* Copyright (c) Madge Networks Ltd 1999
*
* PVCS Version History:
*
*  $Archive:   W:/PVCSARCH/linux/mtok/pcmcia/madgecs.h_v  $
*  $Revision: 1.7 $
*  $Date: 1999/04/13 10:42:39 $
*  $Author: tom $
*  $Modtime:   05 Jun 1998 13:05:22  $
*    
*****************************************************************************/

#ifndef __MTOKCS__
#define __MTOKCS__


#include "driver.h" 
#include "udt/smart.h"

/*----------------------------------------------------------------------------
|
| Version number.
|
| If you (non Madge being) build a customised version of the driver please,
| please define MTOKCS_CUSTOMISED_BY to be a sensible string so that our 
| support people can keep track of what version of code someone with a
| problem is using.
|
|---------------------------------------------------------------------------*/

#define MTOKCS_VERSION         "1.04"
/* #define MTOKCS_CUSTOMISED_BY   "My own version" */


/*----------------------------------------------------------------------------
|
| Update history.
|
|-----------------------------------------------------------------------------
1.04            19/11/1999       TEG (Madge)

                Olicom support addded.

1.03			16/09/1999		 TEG (Madge)

				Removed debug printk's.  Changed WARNING message to ERROR msg.

1.02			14/09/1999				TEG (Madge)

				Cleaned up and changed for new driver interface.

1.01			13/04/1999				TEG (Madge)
	
				Release to test.

1.00.03			08/04/1999				TEG (Madge)

				Changed interface to mtok module for Card Bus Mk 2				

1.00.02									TEG (Madge)

				Internal Version				
				
1.00.01         29/08/1997              PBA (Madge)

                Time begins.
                
                Driver based heavily on skeleton.c written by David Hinds
                (dhinds@allegro.stanford.edu).
                
----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
|
| Product tracking information. This must not be changed.
|
|---------------------------------------------------------------------------*/

#define MTOKCS_PRODUCT_NAME      "Madge Token-Ring PCMCIA Helper for Linux"
#define MTOKCS_PRODUCT_COPYRIGHT "Copyright (c) Madge Networks Ltd 1999"
#define MTOKCS_PRODUCT_PLATFORM  "Linux/x86"


/*----------------------------------------------------------------------------
|
| Some extra constants.
|
|--------------------------------------------------------------------------*/

#define EOK   0

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE 0
#endif


/*----------------------------------------------------------------------------
|
| Driver name.
|
|--------------------------------------------------------------------------*/

#define MTOKCS_NAME "mtok_cs"


/*----------------------------------------------------------------------------
|
| I/O constants.
|
|--------------------------------------------------------------------------*/

#define MTOKCS_IO_RANGE   32
#define MTOKCS_IO_MIN     0x0300
#define MTOKCS_IO_MAX     0x3a20
#define MTOKCS_IO_STEP    0x0020


/*----------------------------------------------------------------------------
|
| Structure used to describe an instance of the driver.
|
|--------------------------------------------------------------------------*/

typedef struct _MTOKCS_INSTANCE
{
    struct _MTOKCS_INSTANCE * next;
    
    /*---------------------------------------------------------------------*/
    
    dev_link_t    link; /* Device linkage structure. */
    dev_node_t    node; /* Device node structure. */
    LINUXDEV *dev;  /* Network device control structure. */

    /*---------------------------------------------------------------------*/
    
    int cs_init;        /* Has Card Services been initialised? */
    int io_init;        /* Has I/O been initialised? */
    int irq_init;       /* Has our IRQ been initialised? */
    int sock_init;      /* Has our socket been initialised? */
    int adap_init;      /* Has the adapter been initialised? */

#ifdef OLICOM_PCMCIA_UDI
    RS_TUPLE_T  olicom_rs_tuple;
    BIA_TUPLE_T olicom_bia_tuple;
#endif    

}
MTOKCS_INSTANCE;


/*----------------------------------------------------------------------------
|
| Functions imported from the TRN driver mtok.o.
|
|---------------------------------------------------------------------------*/

void
mtok_pcmcia_add(unsigned io_base);

void
mtok_pcmcia_remove(unsigned io_base);

struct device *
mtok_create_device(LINUXDEV * dev);

struct device * mtok_get_device(char *name);

void
mtok_destroy_device(LINUXDEV * dev);

int   
mtok_register_device(
	LINUXDEV * dev,
	int             io_base,
	int             pci_number,
	int             irq_number,
	int             ring_speed,
	int             dtr_options, 
	unsigned char * pLAA    
	);
                                

void
mtok_deregister_device(LINUXDEV * dev);

/* mtok_cleanup(void * adap); Really (MTOK_ADAPTER * adap). */

void
mtok_isr(int irq, void * dev_id, struct pt_regs * regs);

int
mtok_probe(LINUXDEV * dev);

void
mtok_cardbus_remove(
    unsigned io_base
        );

void
mtok_cardbus_add(
	unsigned io_base
      );
                
                
        

#endif /* __MTOKCS__ */

/**** End of madgecs.h ******************************************************/
