/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
 * Filename:       mdgndbin.c
 * Author:         Tom Gallagher
 * Date:           13/11/98
 * Description:    Program to extract the two download images from the UDT
 *                 and write them to a file.
 * See Also:       
 *****************************************************************************
 *
 * Copyright (C) 1998 Madge Networks Ltd.
 *
 *****************************************************************************

 *****************************************************************************
 *
 *   $Id: $   
 *    
 *****************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include "driver.h"
#include "mdgbin.h"
#include "knosbin.h"
#include "fmpbin.h"

#if !defined(SMART_UDI) && !defined(KNOSSOS_UDI)
#error No download images needed for current UDT configuration
#endif

#ifdef UDT_NO_IMAGE_HEADERS
#error No image headers
#endif

#ifdef UDT_NO_MVER_STRINGS
#error No mver strings
#endif

#ifdef EMBEDDED_MICRO_CODE
#error Micro code is embedded
#endif

/************************************************************************
 *  Function:   main
 ************************************************************************
 *  Parameters:     argc    - Number of command line parameters
 *                  argv    - Array of pointers to command line parameter.
 *
 *  Description:    Open the output file and then call the two image
 *                  specific routines to write the UDT download images
 *                  to the file.
 *                  The file format output is as follows:-
 *
 *                  KNOSSOS
 *                      Offset  Length   Description
 *                      0000    0002     Length (XXXX)
 *                      0002    0094     Image Header
 *                      0096    XXXX     Image Data
 *                  FMPLUS
 *                     0000     0002     Length (YYYY)
 *                     0002     0094     Image Header
 *                     0096     YYYY     Image Data
 * 
 *  Return Value:   0 = SUCCESS, 1 = FAILURE.
 ***********************************************************************/
int main(int argc,char **argv)
{
    FILE *fptr;

    /*
     * Open the output file for writing in binary mode
     */

#ifdef KNOSSOS_UDI     
    if (NULL==(fptr=fopen(KNOS_FILENAME,"wb"))) 
    {
        fprintf(stderr,"Error opening file %s\n",KNOS_FILENAME);
        return 1;
    }
    printf("Writing - %s\n",KNOS_FILENAME);
    WriteKnossos(fptr);
    fclose(fptr);
#endif

#ifdef SMART_UDI
    if (NULL==(fptr=fopen(FMP_FILENAME,"wb"))) 
    {
        fprintf(stderr,"Error opening file %s\n",FMP_FILENAME);
        return 1;
    }
    printf("Writing - %s\n",FMP_FILENAME);
    WriteFMPlus(fptr);
    fclose(fptr);
#endif

    return 0;
}
 

/************************************************************************
 *  Function:   WriteIt
 ************************************************************************
 *  Parameters:     fptr    - FILE * pointer to the file to write to.
 *                  Head    - Pointer to the start of download image
 *                            including header.
 *                  Len     - The length of the download image excluding
 *                            the image header.
 *
 *  Description:    Writes a file header
 *                  followed immediately by the image header and then
 *                  the image data.
 *                  This is a generic routine for knossos and fmp.
 *  Return Value:   None.
 ***********************************************************************/
void WriteIt(FILE *fptr,void *Data,WORD_T Len)
{
	FILE_HEADER_T Head;
	Head.sig1 = SIG1;
	Head.sig2 = SIG2;
	Head.len = Len;
    /* Signature */
	fwrite(&Head,sizeof(FILE_HEADER_T),1,fptr);
    /* Image Data */
    fwrite(Data,Len,1,fptr);
}

