/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
*
* mpoke.h
*
* Linux pokeout driver header file.
*
* Copyright (c) Madge Networks Ltd 1999
* Company Confidential
*
*****************************************************************************/

#ifndef __MPOKE__
#define __MPOKE__

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*----------------------------------------------------------------------------
|
| To use pokeouts
| ---------------
|
| In your driver/application.
|
| 1.  Call the mpoke_register() function. This takes a single parameter which 
|     is the name of your module. mpoke_register() returns a caller
|     identifier. This identifier should be passed to all other mpoke_???
|     functions. 
|
| 2.  Use the mpoke_printf(), mpoke_pokec(), mpoke_pokestr(), mpoke_pokeb(),
|     mpoke_pokew() and mpoke_pokedw() functions to output pokeout data.
|     moke_printf() is the most flexible function but the others are faster.
|
|     There are also the following handy macros (which assume a global
|     variable called pokeout_handle):
|
|         POKEC(colour, character)
|         POKES(colour, string)
|         POKEB(colour, byte)
|         POKEW(colour, word)                   - 16 bit value
|         POKED(colour, dword)                  - 32 bit value
|         POKESB(colour1, colour2, string, byte)
|         POKESW(colour1, colour2, string, word)
|         POKESD(colour1, colour2, string, dword)
|         POKEn(colour, string, arg1, arg2,..., argn), where n>=1
|
|     In the above macros colour is just the end of the colour name. e.g.
|     RED.
|
| 3.  When you have finished with pokeouts call mpoke_deregister(). 
|
|
| At the Linux console
| --------------------
|
| 1.  Load the mpoke.o pokeout driver. This is done with the insmod command
|     like:
|
|     insmod mpoke.o
|
|     This must be done before you load your driver/application.
|
| 2.  mpoke.o can send pokeouts to one of three places.
|
| 2a. The default is for pokeouts to be written to a cyclic buffer that
|     can be read from a device file with major number 120. Such a device
|     file can be created with the command:
|
|     mknod /dev/mpoke c 120 0
|
|     Pokeouts can then be displayed on the console by using cat, e.g.:
|
|     cat /dev/mpoke
|
|     Pokeouts can be displayed on the console and written to a file using
|     cat and tee, e.g.:
|
|     cat /dev/mpoke | tee /tmp/pokeouts
|
|     The major drawback of this scheme is that the user mode program that
|     displays the pokeouts (e.g. cat) must be working. i.e. if a driver 
|     hangs in an interrupt handler you won't see any pokeouts.
|
| 2b. If mpoke.o is started with the parameter "console=1" then 
|     pokeouts will be sent to the current console. e.g.:
|
|     insmod mpoke.o console=1
|
|     This method is faster and more robust than the default method but still 
|     uses the Linux console driver to do the output. All pokeouts will be 
|     monochrome using this method.
|
| 2c. If mpoke.o is started with the parameter "rawvga=1" then pokeouts
|     will be written directly to VGA memory. e.g.:
|
|     insmod mpoke.o rawvga=1
|
|     A side effect of this method is that normal console handling is messed
|     up. The Linux VGA console driver does some clever re-mapping of VGA
|     memory to optimise scrolling and for virtual consoles. Therefore it
|     is hard to workout which extent of VGA memory is visible. To avoid
|     this problem, mpoke.o remaps the visible extent of VGA memory to start
|     at its base (i.e. 0xb8000). As a consequence when mpoke.o is running in
|     rawvga mode the normal console is not visible. To get back to the
|     normal console you should switch to a different virtual console (by
|     pressing CTRL-ALT-FUNCTION_KEY). The pokeout screen will be restored
|     after 10 seconds.
|  
| 3.  To force all pokeouts to be in monochrome use the parameter "colour=0".
|     You may wish to do this when sending pokeouts to a file as in the
|     output mode that uses a device file (2a above) colour is implemented
|     with ANSI escape sequences.
|   
| 4.  To remove mpoke.o use the rmmod command like:
|
|     rmmod mpoke
|
|---------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
|
| Pokeout colours. Specify one colour and optionally a white background by
| ORing in the MPOKE_COLOUR_WHITE_BGRND value.
|
|---------------------------------------------------------------------------*/

#define MPOKE_WHITE           0
#define MPOKE_BLACK           1
#define MPOKE_BLUE            2
#define MPOKE_GREEN           3
#define MPOKE_CYAN            4
#define MPOKE_RED             5
#define MPOKE_MAGENTA         6
#define MPOKE_YELLOW          7
#define MPOKE_BRIGHT_WHITE    8
#define MPOKE_BRIGHT_BLACK    9
#define MPOKE_BRIGHT_BLUE     10
#define MPOKE_BRIGHT_GREEN    11
#define MPOKE_BRIGHT_CYAN     12
#define MPOKE_BRIGHT_RED      13
#define MPOKE_BRIGHT_MAGENTA  14
#define MPOKE_BRIGHT_YELLOW   15

#define MPOKE_WHITE_BGRND     16


/*----------------------------------------------------------------------------
|
| Function    - mpoke_register
|
| Parameters  - name -> Name of the caller registering.
|
| Purpose     - Called to register with the pokeout driver. This function
|               must be called before any other mpoke functions.
|
| Returns     - A caller identifer to be passed into other mpoke functions.
|
|---------------------------------------------------------------------------*/

int
mpoke_register(
    char * name);


/*----------------------------------------------------------------------------
|
| Function    - mpoke_deregister
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|
| Purpose     - Called to deregister with the pokeout driver.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void
mpoke_deregister(
    int ident);


/*----------------------------------------------------------------------------
|
| Function    - mpoke_printf
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               fmt    -> A printf style format string.
|               ...    -> A variable number of arguments.
|
| Purpose     - Do a pokeout in the style of printf.
|
| Returns     - We don't really return anything. mpoke_printf() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_printf(
    int    ident,
    int    colour,
    char * fmt,
    ...);
    

/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokec
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               c      -> A character to pokeout.
|
| Purpose     - Pokeout a character.
|
| Returns     - We don't really return anything. mpoke_pokec() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokec(
    int ident,
    int colour,
    int c);


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokestr
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               str    -> A string to pokeout.
|
| Purpose     - Pokeout a string.
|
| Returns     - We don't really return anything. mpoke_pokestr() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokestr(
    int    ident,
    int    colour,
    char * str);


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokeb
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               b      -> An 8bit byte to pokeout.
|
| Purpose     - Pokeout an 8 bit byte in hex.
|
| Returns     - We don't really return anything. mpoke_pokeb() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokeb(
    int ident,
    int colour,
    int b);


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokew
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               w      -> A 16bit word to pokeout.
|
| Purpose     - Pokeout a 16 bit word in hex.
|
| Returns     - We don't really return anything. mpoke_pokew() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokew(
    int      ident,
    int      colour,
    unsigned w);


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokedw
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               dw     -> A 32bit dword to pokeout.
|
| Purpose     - Pokeout a 32 bit dword in hex.
|
| Returns     - We don't really return anything. mpoke_pokedw() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokedw(
    int           ident,
    int           colour,
    unsigned long dw);

/*----------------------------------------------------------------------------
|
| Pokeout macros.
|
|---------------------------------------------------------------------------*/
#if 0

/*
 * Pokeout char.
 */

#define POKEC(_colour, _character)                                           \
        mpoke_pokec(pokeout_handle, MPOKE_ ## _colour, _character)

/*
 * Pokeout string.
 */

#define POKES(_colour, _string)                                              \
        mpoke_pokestr(pokeout_handle, MPOKE_ ## _colour, _string)

/*
 * Pokeout value.
 */

#define POKEB(_colour, _byte)                                                \
        mpoke_pokeb(pokeout_handle, MPOKE_ ## _colour, _byte)

#define POKEW(_colour, _word)                                                \
        mpoke_pokew(pokeout_handle, MPOKE_ ## _colour, _word)

#define POKED(_colour, _dword)                                               \
        mpoke_pokedw(pokeout_handle, MPOKE_ ## _colour, _dword)

/*
 * String and value (in different colours).
 */

#define POKESB(_colour1, _colour2, _string, _byte)                           \
        mpoke_pokestr(pokeout_handle, MPOKE_ ## _colour1, _string);          \
        mpoke_pokeb(pokeout_handle, MPOKE_ ## _colour2, _byte)

#define POKESW(_colour1, _colour2, _string, _word)                           \
        mpoke_pokestr(pokeout_handle, MPOKE_ ## _colour1, _string);          \
        mpoke_pokew(pokeout_handle, MPOKE_ ## _colour2, _word)

#define POKESD(_colour1, _colour2, _string, _dword)                          \
        mpoke_pokestr(pokeout_handle, MPOKE_ ## _colour1, _string);          \
        mpoke_pokedw(pokeout_handle, MPOKE_ ## _colour2, _dword)

/*
 * Printf-esque macros.
 */

#define POKE1(_colour, _string, _arg1)                                       \
        mpoke_printf(pokeout_handle, MPOKE_ ## _colour, _string, _arg1)

#define POKE2(_colour, _string, _arg1,                                       \
                                _arg2)                                       \
        mpoke_printf(pokeout_handle, MPOKE_ ## _colour, _string, _arg1,      \
                                                                  _arg2)

#define POKE3(_colour, _string, _arg1,                                       \
                                _arg2,                                       \
                                _arg3)                                       \
        mpoke_printf(pokeout_handle, MPOKE_ ## _colour, _string, _arg1,      \
                                                                  _arg2,     \
                                                                  _arg3)

#define POKE4(_colour, _string, _arg1,                                       \
                                _arg2,                                       \
                                _arg3,                                       \
                                _arg4)                                       \
        mpoke_printf(pokeout_handle, MPOKE_ ## _colour, _string, _arg1,      \
                                                                  _arg2,     \
                                                                  _arg3,     \
                                                                  _arg4)

#define POKE5(_colour, _string, _arg1,                                       \
                                _arg2,                                       \
                                _arg3,                                       \
                                _arg4,                                       \
                                _arg5)                                       \
        mpoke_printf(pokeout_handle, MPOKE_ ## _colour, _string, _arg1,      \
                                                                  _arg2,     \
                                                                  _arg3,     \
                                                                  _arg4,     \
                                                                  _arg5)

#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */
    
#endif /* not __MPOKE__ */    

/**** End of mpoke.h ******************************************************/
