/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
*
* mver.c
*
* Utility to extract and set product tracking strings in binaries.
*
* Copyright (c) Madge Networks Ltd 1999
*
* $Id $
*
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


/*----------------------------------------------------------------------------
|
| Version number.
|
|---------------------------------------------------------------------------*/

#define VERSION_NUMBER "1.02"

/* #define IS_DEVELOPMENT */ /* Define if development alpha version. */
/* #define IS_RESTRICTED  */ /* Define if restricted alpha version. */

#define LSS_VERSION     /* "LSS 6.1(1)" */

#define COPYRIGHT       "Copyright (c) Madge Networks Ltd 1999"


/*----------------------------------------------------------------------------
|
| Update history.
|
|-----------------------------------------------------------------------------

1.03		09/09/1999		TEG

			Changed return type of main to int.

1.02		13/04/1999		TEG

			Added PLATFORM_NAME defines for Linux

1.01        25/11/1997      PBA

            Pre-released for LSS 6.1.
            
1.00.01     24/04/1996      PBA

            First version.

----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
|
| Notes on conditional compilation.
|
|-----------------------------------------------------------------------------
|
| The default is to build the complete full featured version of vertool.
| If BUILD_DISPLAY is defined then a version will be build that can only
| be used to display the product tracking strings. If BUILD_RESTRICTED
| is defined then a version is built that only allows display and validation 
| of product tracking strings along with setting of the RESTRICTED string.
|
|---------------------------------------------------------------------------*/

#if !defined(BUILD_DISPLAY) && !defined(BUILD_RESTRICTED)
#define BUILD_FULL 1
#endif

            
/*---------------------------------------------------------------------------
|
| Work out what platform we're on.
|
|--------------------------------------------------------------------------*/

#if defined(_MSDOS_)

#define PLATFORM_NAME "MSDOS"

#elif defined(_WIN32_ALPHA_)

#define PLATFORM_NAME "WIN32/ALPHA"

#elif defined(_WIN32_MIPS_)

#define PLATFORM_NAME "WIN32/MIPS"

#elif defined(_WIN32_PPC_)

#define PLATFORM_NAME "WIN32/PPC"

#elif defined(_WIN32_I386_)

#define PLATFORM_NAME "WIN32/x86"

#elif defined(_AIX41x_)

#define PLATFORM_NAME "AIX4.1.x"

#elif defined(_AIX32x_)

#define PLATFORM_NAME "AIX3.2.x"

#elif defined(_SCO3_)

#define PLATFORM_NAME "SCO3"

#elif defined(_SCO5_)

#define PLATFORM_NAME "SCO5"

#elif defined(_SOLARIS24_)

#define PLATFORM_NAME "Solaris 2.4/x86"

#elif defined(_SOLARIS25_)

#ifdef i386

#define PLATFORM_NAME "Solaris 2.5x/x86"

#else

#define PLATFORM_NAME "Solaris 2.5x/Sparc"

#endif

#elif defined(_SOLARIS26_)

#ifdef i386

#define PLATFORM_NAME "Solaris 2.5x/2.6/x86"

#else

#define PLATFORM_NAME "Solaris 2.5x/2.6/Sparc"

#endif

#elif defined(__linux__)

#if  defined(i386)

#define PLATFORM_NAME "Linux x86"

#elif  defined(alpha)

#define PLATFORM_NAME "Linux Alpha"

#else

#define PLATFORM_NAME "Linux"

#endif

#endif


/*---------------------------------------------------------------------------
|
| Macro to create the special "MVER" product tracking strings.
|
|--------------------------------------------------------------------------*/

#define MVER_PREFIX_SIZE  8

#if defined(BUILD_DISPLAY)

static char mver_title[] = 
    "VeRsIoN=Madge Product Tracking String Display Utility"; 

#elif defined(BUILD_RESTRICTED)

static char mver_title[] = 
    "VeRsIoN=Madge Restricted Product Tracking String Maintenance Utility"; 

#else

static char mver_title[] = 
    "VeRsIoN=Madge Product Tracking String Maintenance Utility"; 

#endif
    
static char mver_copyright[] = 
    "VeRsIoN=" COPYRIGHT; 

static char mver_information[] =
    "VeRsIoN=INFO: mver.txt";

#ifdef PLATFORM_NAME
    
static char mver_platform[] = 
    "VeRsIoN=PLATFORM: " PLATFORM_NAME;

#endif
    
#if defined(IS_DEVELOPMENT)

static char mver_version[61] =
    "VeRsIoN=VERSION: v" VERSION_NUMBER "D " LSS_VERSION;

static char mver_restricted[61] = 
    "VeRsIoN=RESTRICTED: DEVELOPMENT ONLY";

#elif defined(IS_RESTRICTED)

static char mver_version[61] =
    "VeRsIoN=VERSION: v" VERSION_NUMBER " " LSS_VERSION;

static char mver_restricted[61] = 
    "VeRsIoN=RESTRICTED: RESTRICTED USE";

#else

static char mver_version[61] =
    "VeRsIoN=VERSION: v" VERSION_NUMBER " " LSS_VERSION;

static char * mver_restricted = NULL;

#endif    


/*----------------------------------------------------------------------------
|
| Some extra types.
|
|---------------------------------------------------------------------------*/

#define BOOLEAN   int
#define TRUE      1
#define FALSE     0

#define WORD      unsigned short 
#define BYTE      unsigned char


/*----------------------------------------------------------------------------
|
| CRC contants.
|
|---------------------------------------------------------------------------*/

#define CRC_GEN        0x1021     /* The CRC generator polynomial. */
#define CRC_MASK       "DGDS"     /* The CRC mask. */
#define CRC_MASK_SIZE  4          
#define CRC_UNINIT     "XXXX"     /* Unitialised CRC string. */


/*----------------------------------------------------------------------------
|
| Miscellaneous constants.
|
|---------------------------------------------------------------------------*/

#define MAX_NAME_SIZE      128
#define MAX_STRING_SIZE    256


/*----------------------------------------------------------------------------
|
| String termination test macro.
|
|---------------------------------------------------------------------------*/

#define IS_TERM(_x)  ((_x) == '@' || (_x) == '\0')


/*----------------------------------------------------------------------------
|
| Program name.
|
|---------------------------------------------------------------------------*/

char * prog_name;


/*----------------------------------------------------------------------------
|
| Macros and variables for error handling.
|
|---------------------------------------------------------------------------*/

int _error_code = 0;

#define RETURN(_x)   return (_error_code = (_x))
#define RETURNX()    return (_error_code)


/*----------------------------------------------------------------------------
|
| Error codes and explinations.
|
|---------------------------------------------------------------------------*/

#define E_OK          0
#define E_FILE        1   /* Cannot access a file. */
#define E_MEMORY      2   /* Cannot allocate memory. */
#define E_NO_CRC      3   /* No CRC found in image.           */
#define E_CRC_SIZE    4   /* The CRC found is the wrong size. */
#define E_UNINIT_CRC  5   /* The CRC found has not been initialised. */
#define E_INVALID_CRC 6   /* The CRC found contains an invalid value. */
#define E_UNTERM_STR  7   /* Unterminated string found in file. */
#define E_STR_SIZE    8   /* String is too long. */
#define E_NO_STR      9   /* String not found. */
#define E_BAD_ARGS    10  /* Bad command line arguments. */
#define E_STR_UNKNOWN 11  /* Unknown string type. */
#define E_STR_COUNT   12  /* A string occurs the wrong number of times */


/*----------------------------------------------------------------------------
|
| The "magic" character strings used in binaries to mark that a 
| product tracking string or CRC follows. We store the first character
| separately so that we don't confuse ouselves if we look at ourselves.
|
|---------------------------------------------------------------------------*/

char mver_marker[] = "_eRsIoN=";
char crc_marker[]  = "_aDgE=";

#define MVER_MARKER_FIRST  'V'
#define CRC_MARKER_FIRST   'M'


/*----------------------------------------------------------------------------
|
| Table of the product tracking strings.
|
|---------------------------------------------------------------------------*/

typedef struct
{
    char    * key;      /* The key prefixing the string in the binary. */
    char    * display;  /* The prefix to use when displaying. */
    int       min_len;  /* Minimum length allowed. */
    int       max_len;  /* Maximum length allowed. */
    int       min_num;  /* Minimum number allowed. */
    int       max_num;  /* Maximum number allowed. */
    int       count;    /* Some workspace for validate_strings. */
}
STRING_OBJ;

STRING_OBJ string_table[] = 
{
    {"VERSION",    "Version     ",   1,  60,   1,   1},
    {"PRODUCT-ID", "Product-ID  ",  30,  30,   1,   3},
    {"PLATFORM",   "Platform    ",   1,  60,   0,   1},
    {"RESTRICTED", "Restricted  ",   1,  60,   0,   5},
    {"INFO",       "Info        ",   1, 200,   1,   1},
    {"MAC",        "MAC         ",   1,  60,   0,   1},
    {NULL,          NULL         ,   0,   0,   0,   0}
};


#define RESTRICTED_KEY  "RESTRICTED"     /* Name of the restricted key.*/


/*----------------------------------------------------------------------------
|
| Structure to hold details of an in-core file. 
|
|---------------------------------------------------------------------------*/

typedef struct
{
    char     name[MAX_NAME_SIZE];
    char   * image;
    size_t   size;
}
FILE_OBJ;


/*----------------------------------------------------------------------------
|
| Operational modes.
|
|---------------------------------------------------------------------------*/

#define MODE_DISPLAY          0   /* Display the product tracking strings. */
#define MODE_SET_CRC          1   /* Set the CRC. */
#define MODE_COPY_STRING      2   /* Copy a string from on file to another. */
#define MODE_PATCH_STRING     3   /* Patch a string. */
#define MODE_PATCH_RESTRICTED 4   /* Patch the RESTRICTED string. */
#define MODE_VALIDATE         5   /* Check the format of the strings. */


/*----------------------------------------------------------------------------
|
| Function   - load_file
|
| Parameters - name -> Path/name of the file to be loaded.
|              file -> Pointer to a FILE_OBJ to hold details of the
|                      loaded file.
|
| Purpose    - Load a file into an in-core buffer allocated by this
|              function.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/
        
int
load_file(
    char     * name,
    FILE_OBJ * file
    )
{
    FILE          * inf;
    unsigned long   size;
    size_t          left;
    size_t          n;
    char          * ptr;
    
    /*
     * Open the file to load.
     */
     
    if ((inf = fopen(name, "rb")) == NULL)
    {
        fprintf(stderr, "%s: cannot read file '%s'\n", prog_name, name);
        RETURN(E_FILE);
    }
    
    strcpy(file->name, name);
    
    /*
     * Work out the size of the file.
     */
     
    fseek(inf, 0L, SEEK_END);
    size = (unsigned long) ftell(inf);
    fseek(inf, 0L, SEEK_SET);
    
    /*
     * Check that we can allocate a memory block large enough
     * in this architecture.
     */
     
    if (size > (size_t) size)
    {
        fprintf(stderr, "%s: cannot allocate large enough memory block\n", prog_name);
        fclose(inf);
        RETURN(E_MEMORY);
    }     
    
    file->size = (size_t) size;
    
    /*
     * Allocate memory for the image.
     */
     
    if ((file->image = (char *) malloc(file->size)) == NULL)
    {
        fclose(inf);
        fprintf(stderr, "%s: cannot allocate enough memory\n", prog_name);
        RETURN(E_MEMORY);
    }     
    
    /*
     * Read the image into the memory allocated.
     */
     
    left = file->size;
    ptr  = file->image; 
    
    while (left > 0)
    {
        n = (left > 16384) ? 16384 : left;
                 
        if (fread(ptr, 1, n, inf) != n)
        {
            fclose(inf);
            fprintf(stderr, "%s: cannot read file '%s'\n", prog_name, name);
            RETURN(E_FILE);
        }
        
        left -= n;
        ptr  += n;
    }
   
    fclose(inf);
    
    /*
     * And return success.
     */
     
    RETURN(E_OK);
}

                       
/*----------------------------------------------------------------------------
|
| Function   - save_file
|
| Parameters - file -> Pointer to a FILE_OBJ to hold details of the
|                      loaded file.
|
| Purpose    - Save the held in the in-core buffer specified by *file.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/
        
int
save_file(
    FILE_OBJ * file
    )
{
    FILE          * outf;
    size_t          left;
    size_t          n;
    char          * ptr;
    
    /*
     * Open the file to be written.
     */
     
    if ((outf = fopen(file->name, "r+b")) == NULL)
    {
        fprintf(stderr, "%s: cannot write file '%s'\n", prog_name, file->name);
        RETURN(E_FILE);
    }
    
    /*
     * Write the in-core image to file.
     */
     
    left = file->size;
    ptr  = file->image; 
    
    while (left > 0)
    {
        n = (left > 16384) ? 16384 : left;
                 
        if (fwrite(ptr, 1, n, outf) != n)
        {
            fclose(outf);
            fprintf(stderr, "%s: cannot write file '%s'\n", prog_name, file->name);
            RETURN(E_FILE);
        }
        
        left -= n;
        ptr  += n;
    }
   
    fclose(outf);
    
    /*
     * And return success.
     */
     
    RETURN(E_OK);
}

                       
/*----------------------------------------------------------------------------
|
| Function   - free_file
|
| Parameters - file -> Pointer to a FILE_OBJ to hold details of the
|                      loaded file.
|
| Purpose    - Free the in-core buffer specified by *file.
|
| Returns    - Nothing.
|
|---------------------------------------------------------------------------*/
        
void
free_file(
    FILE_OBJ * file
    )
{
    free(file->image);
}

                           
/*----------------------------------------------------------------------------
|
| Function   - find_string
|
| Parameters - file   -> Pointer to a FILE_OBJ desribing the in-core
|                        image to be searched.
|              prefix -> Prefix for which to search.
|              offset -> Pointer to the offset from which to start
|                        searching and the string offset on successfull
|                        return.
|
| Purpose    - Search the in-core image described by *file for a string
|              starting with the prefix prefix. To start searching at
|              the beginning of the image *offset should be 0. Otherwise
|              it should be the value returned from the previous call. On
|              successfull return *offset will be the offset of the start
|              of the string following the prefix.
|
| Returns    - TRUE if a string is found or FALSE if not.
|
|---------------------------------------------------------------------------*/

BOOLEAN
find_string(
    FILE_OBJ * file,
    char     * prefix,
    size_t   * offset
    )
{
    size_t offs;
    size_t len;
    
    offs = *offset;
    len  = strlen(prefix);
        
    while (offs < file->size - len)
    {
        if (file->image[offs    ] == prefix[0] &&
            file->image[offs + 1] == prefix[1])
        {
            if (memcmp(&file->image[offs], prefix, len) == 0)
            {
                *offset = offs + len;
                return TRUE;
            }
        }
        
        offs++;
    }
    
    return FALSE;                     
}


/*----------------------------------------------------------------------------
|
| Function   - identify_key
|
| Parameters - file   -> Pointer to a FILE_OBJ describing the in-core
|                        image containing the key.
|              offset -> Offset of the key in the in-core image.
|
| Purpose    - Look in the in-core image specified starting at offset
|              to see if there is an "MVER" string key present. A key
|              is anything of the form "XXX: ". Where XXX is a sequence
|              of none whitespace characters.
|
| Returns    - Length of the key if there is one present or 0 if not.
|
|---------------------------------------------------------------------------*/

size_t
identify_key(
    FILE_OBJ * file,
    size_t     offset
    )
{
    size_t len;
    
    len = 1;
    
    for (;;)
    {
        /*
         * If we run out of string then we've failed to find a key.
         */
         
        if (offset >= file->size || IS_TERM(file->image[offset]))
        {
            return 0;
        }
    
        /*
         * If hit a colon then we found a key.
         */
         
        if (file->image[offset] == ':')
        {
            return len - 1;
        }
            
        /*
         * If we hit a space character then we are are not inside a key.
         */
         
        if (isspace(file->image[offset]))
        {
            return 0;
        }
        
        /*
         * We are still inside a key so check the next character.
         */
         
        offset++;
        len++;
    }
}


/*----------------------------------------------------------------------------
|
| Function   - lookup_key
|
| Parameters - key -> The key to lookup.
|
| Purpose    - Lookup a key in the string table.
|
| Returns    - A pointer to the string table entry for the key or
|              NULL if the key is not in the table.
|
|---------------------------------------------------------------------------*/


STRING_OBJ *
lookup_key(
    char * key
    )
{
    STRING_OBJ * str;
    
    for (str = string_table; str->key != NULL; str++)
    {
        if (strcmp(str->key, key) == 0)
        {
            return str;
        }
    }
    
    return NULL;
}

                             
/*----------------------------------------------------------------------------
|
| Function   - extract_string
|
| Parameters - file     -> Pointer to a FILE_OBJ describing the in-core
|                          image to be searched.
|              offset   -> Offset from which to start extracting the string.
|              buffer   -> Buffer for the extracted string.
|              buf_size -> Size of the buffer.
|
| Purpose    - Extract the string that starts at offset offset in the
|              in-core image described by file. Leading white space is
|              removed.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/

int
extract_string(
    FILE_OBJ * file,
    size_t     offset,
    char     * buffer,
    size_t     buf_size
    )
{
    size_t  len;
    BOOLEAN in_space;
    
    len      = 0;
    in_space = TRUE;
    
    while (!IS_TERM(file->image[offset]))
    {
        if (offset >= file->size - 1)
        {
            fprintf(stderr, "%s: unterminated string found\n", prog_name);
            RETURN(E_UNTERM_STR);
        }
           
        if (len >= buf_size - 1)
        {
            fprintf(stderr, "%s: string too long\n", prog_name);
            RETURN(E_STR_SIZE);
        }
            
        if (!isspace(file->image[offset]))
        {
            in_space = FALSE;
        }
        
        if (!in_space)
        {        
            buffer[len] = file->image[offset];
            len++;
        }
            
        offset++;
    }
    
    buffer[len] = '\0';
    
    RETURN(E_OK);
}


/*----------------------------------------------------------------------------
|
| Function   - find_crc
|
| Parameters - file   -> Pointer to a FILE_OBJ desribing an in-core
|                        image.
|              offset -> Pointer to a holder for the offset of the CRC in
|                        the image.
|              crc    -> Pointer to a holder for the CRC found.
|
| Purpose    - Find and validate the format of the CRC in an in-core
|              image.
|
| Returns    - An error code. 
|
|                  E_OK          means the CRC was found and is valid.
|                  E_NO_CRC      means no CRC was found.
|                  E_CRC_SIZE    means the CRC is an invalid size.
|                  E_UNINIT_CRC  means the CRC has not been initialised.
|                  E_INVALID_CRC means the CRC contains an invalid character
|                                but is the correct size.
|
|---------------------------------------------------------------------------*/

int
find_crc(
    FILE_OBJ * file,
    size_t   * offset,
    WORD     * crc
    )
{
    char buf[MAX_STRING_SIZE];
    int  i;
    
    /*
     * Search the in-core image for the CRC marker and if found extract
     * the CRC string from the image.
     */
     
    *offset = 0;
     
    if (!find_string(file, crc_marker, offset))
    {
        RETURN(E_NO_CRC);
    }
    
    if (extract_string(file, *offset, buf, MAX_STRING_SIZE) != E_OK)
    {
        RETURN(E_CRC_SIZE);
    }    

    /*
     * The CRC string should be exactly 4 characters long.
     */
     
    if (strlen(buf) != 4)
    {
        RETURN(E_CRC_SIZE);
    }
    
    /*
     * The CRC may be unitialised in which case the CRC string will
     * have the unitialised value of "XXXX".
     */
     
    if (strcmp(buf, CRC_UNINIT) == 0)
    {
        RETURN(E_UNINIT_CRC);
    }
    
    /*
     * The CRC must consist of 4 hex digits.
     */
     
    if (!isxdigit(buf[0]) || !isxdigit(buf[1]) || 
        !isxdigit(buf[2]) || !isxdigit(buf[3]))
    {
        RETURN(E_INVALID_CRC);
    }
    
    /*
     * Otherwise it must be Ok so extract the value.
     */

    *crc = 0;
    
    for (i = 0; i < 4; i++)
    {
        *crc *= 16;
        
        if (buf[i] >= '0' && buf[i] <= '9')
        {
            *crc += buf[i] - '0';
        }
        else if (buf[i] >= 'A' && buf[i] <= 'F')
        {
            *crc += buf[i] - 'A' + 10;
        }
        else
        {
            *crc += buf[i] - 'a' + 10;
        }
    }
                    
    RETURN(E_OK);
}
                   
                   
/*----------------------------------------------------------------------------
|
| Function   - generate_crc
|
| Parameters - file       -> Pointer to a FILE_OBJ desribing an in-core
|                            image.
|              crc_offset -> The offset in the image of an existing CRC.
|              mask       -> A byte mask to use in place of the existing
|                            CRC.
|              mask_len   -> The length of mask.
|
| Purpose    - Generate a CRC for the in-core image specified.
|
| Returns    - The CRC.
|
|---------------------------------------------------------------------------*/

WORD 
generate_crc(
    FILE_OBJ * file, 
    size_t     crc_offset, 
    char     * mask, 
    int        mask_len
    )
{
    static WORD      crc_tab[256];
    static BOOLEAN   init = FALSE;
    
    WORD             ans;
    WORD             cword;
    BYTE           * mask_ptr;
    int              i; 
    int              j; 
    size_t           k;
    BYTE             temp;

    /*
     * Initialise the CRC lookup table if this has not already been done.
     */
  
    if (!init)
    {
        init = TRUE;
        
        for (j = 0; j < 256; j++)
        {
            ans = (WORD) (j << 8);
        
            for (i = 0; i < 8; i++)
            {
                if ((ans & 0x8000) != 0) 
                {
                    ans = (ans << 1) ^ CRC_GEN;
                }
                else    
                {
                    ans <<= 1;
                }
            }
        
            crc_tab[j] = ans;
        }
    }    

    /*
     * Generate the CRC of the in-core image.
     */
     
    cword    = 0;
    mask_ptr = (BYTE *) mask;
     
    for (k = 0; k < file->size; k++)
    {
        temp = (BYTE) file->image[k]; 
        
        /*
         * If we're inside the CRC that's already in the image
         * then use the fixed mask value rather than the byte
         * from the image.
         */
         
        if (mask_len > 0 && k >= crc_offset)
        {
            cword=crc_tab[*mask_ptr ^ (cword >> 8)] ^((0xff & cword) <<8);
            mask_ptr++;
            mask_len--;
        }
        
        /*
         * Otherwise use the byte from the image.
         */
         
        else
        {
            cword = crc_tab[temp ^ (cword >> 8)] ^ ((0xff & cword) << 8);
        }
    }
    
    return cword;
}


#if defined(BUILD_FULL) || defined(BUILD_RESTRICTED)

/*----------------------------------------------------------------------------
|
| Function   - set_crc
|
| Parameters - name -> Name/path of the file to examine or NULL if
|                      file describes an already loaded image.
|              file -> Specifies an in-core file image if name is NULL.
|
| Purpose    - Set the CRC in a file. If name is not NULL then this function
|              loads and saves the file. Otherwise it just sets the CRC
|              in the image specified by file.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/

int
set_crc(
    char     * name,
    FILE_OBJ * file
    )
{
    FILE_OBJ ifile;
    WORD     crc;
    size_t   offset;
    char     buf[8];
    int      ret_code;
    
    /*
     * If we've been told to load the file ourselves then do so.
     */
     
    if (name != NULL)
    {
        if (load_file(name, &ifile) != E_OK)
        {
            RETURNX();
        }
    }
    
    /*
     * Otherwise use the in-core image specified.
     */
     
    else
    {
        ifile = *file;
    }
         
    /*
     * Find the CRC in the in-core image of the file.
     */
     
    ret_code = find_crc(&ifile, &offset, &crc);
    
    /*
     * Abort with an error message if the CRC string was not found or is
     * not of the correct size.
     */
     
    if (ret_code == E_NO_CRC)
    {
        fprintf(stderr, "%s: the CRC was not found\n", prog_name);
    }
    
    else if (ret_code == E_CRC_SIZE)
    {
        fprintf(stderr, "%s: The CRC has an invalid format\n", prog_name);
    }
    
    else if (ret_code != E_OK && ret_code != E_UNINIT_CRC)
    {
        fprintf(stderr, "%s: the CRC cannot be set\n", prog_name);
    }
 
    /*
     * If we cannot proceed the unload the file if we lodade it
     * and abort.
     */
     
    if (ret_code != E_OK && ret_code != E_UNINIT_CRC)
    {
        if (name != NULL)
        {
            free_file(&ifile);
        }    
        
        RETURN(ret_code);
    }
        
    /*
     * Calculate the CRC.
     */
     
    crc = generate_crc(&ifile, offset, CRC_MASK, CRC_MASK_SIZE);
    
    /*
     * Convert the CRC into a 4 digit hex value and write it into the
     * in-core image. We don't need to worry about the string terminator
     * as we are only replacing the string contents and not the terminator.
     */
     
    sprintf(buf, "%04X", crc);
    memcpy(&ifile.image[offset], buf, 4);
    
    /*
     * If we were told to load the file ourselves then write the file.
     */
     
    if (name != NULL)
    {
        save_file(&ifile);
        free_file(&ifile);
        RETURNX();
    }
    
    RETURN(E_OK);
}

#endif

                
/*----------------------------------------------------------------------------
|
| Function   - display_strings
|
| Parameters - name -> Name/path of the file to examine.
|
| Purpose    - Display the product tracking strings in the file specified.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/

int
display_strings(
    char * name
    )
{
    FILE_OBJ     file;
    STRING_OBJ * str;
    size_t       offset;
    char         buf[MAX_STRING_SIZE];
    char         prefix[MAX_STRING_SIZE];
    WORD         crc;
    int          ret_code;

    /*
     * Load the image file into core.
     */
     
    if (load_file(name, &file) != E_OK)
    {
        RETURNX();
    }
             
    /*
     * First we need to display any title strings. These are
     * strings prefixed with "VeRsIoN=" but not followed by
     * a key. Life would be easier if title strings had
     * a key so we could do all of this processing on go.
     */
     
    offset = 0;
    
    while (find_string(&file, mver_marker, &offset))
    {
        /*
         * We've found a string in the image so now check that
         * it is a title string. If it is then extract it and
         * display it.
         */
         
        if (identify_key(&file, offset) == 0)
        {
            if (extract_string(&file, offset, buf, MAX_STRING_SIZE) == E_OK)
            {
                if (strlen(buf) != 0)
                {
                    printf("%s\n", buf);
                }    
            }    
        }
    }        

    printf("\n");
                
    /*
     * Now we can look for all of the strings that have the form
     * "VeRsIoN=<KEY>:". We'll check each entry in the string table in
     * turn so we display the strings in the correct order.
     */
     
    for (str = string_table; str->key != NULL; str++)
    {
        offset = 0;
        
        strcpy(prefix, mver_marker);
        strcat(prefix, str->key);
        strcat(prefix, ":");
        
        while (find_string(&file, prefix, &offset))
        {
            /*
             * We've found a string in the image so display it.
             */
         
            if (extract_string(&file, offset, buf, MAX_STRING_SIZE) == E_OK)
            {
                printf("%s%s\n", str->display, buf);
            }    
        }
    }
    
    /*
     * Now look for a CRC.
     */
     
    ret_code = find_crc(&file, &offset, &crc);
    
    /*
     * If the CRC was found Ok then calculate what the CRC should be and
     * compare.
     */
     
    if (ret_code == E_OK)
    {
        if (generate_crc(&file, offset, CRC_MASK, CRC_MASK_SIZE) != crc)
        {
            ret_code = E_INVALID_CRC;
        }
    }
    
    /*
     * Otherwise display an appropriate message for the return code.
     */
     
    if (ret_code == E_NO_CRC)
    {
        /* printf("\nThe CRC was not found\n"); */
    }
    
    else if (ret_code == E_CRC_SIZE)
    {
        printf("\nThe CRC has an invalid format\n");
    }
    
    else if (ret_code == E_UNINIT_CRC)
    {
        printf("\nThe CRC has not been set\n");
    }
    
    else if (ret_code != E_OK)
    {
        printf("\nInvalid CRC - this file may be corrupt\n");
    }
        
    /*
     * Unload the image.
     */
     
    free_file(&file);
    
    RETURN(ret_code);               
}


#if defined(BUILD_FULL) || defined(BUILD_RESTRICTED)

/*----------------------------------------------------------------------------
|
| Function   - validate_strings
|
| Parameters - name -> Name/path of the file to examine.
|
| Purpose    - Validate the product tracking strings in the file specified.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/

int
validate_strings(
    char * name
    )
{
    FILE_OBJ     file;
    STRING_OBJ * str;
    size_t       offset;
    char         buf[MAX_STRING_SIZE];
    WORD         crc;
    int          len;
    int          temp_ret_code;
    int          ret_code;

    /*
     * Load the image file into core.
     */
     
    if (load_file(name, &file) != E_OK)
    {
        RETURNX();
    }

    /*
     * Zero the string occurrance counts.
     */
     
    for (str = string_table; str->key != NULL; str++)
    {
        str->count = 0;
    }
                      
    /*
     * Go through the image looking for all strings that start with
     * something that has the correct format for a key.
     */
    
    ret_code = E_OK; 
    offset   = 0;
    
    while (find_string(&file, mver_marker, &offset))
    {
        if ((len = identify_key(&file, offset)) == 0)
        {
            continue;
        }
            
        /*
         * We have something that should be a key. Extract the 
         * whole string. Abandon this entry if the string cannot
         * be extracted.
         */
             
        if (extract_string(&file, offset, buf, MAX_STRING_SIZE) != E_OK)
        {
            ret_code = E_STR_SIZE;
            continue;
        }    
            
        /*
         * Patch the ':' with a null and check that the key is
         * valid.
         */
             
        buf[len] = '\0';
           
        if ((str = lookup_key(buf)) == NULL)
        {
            ret_code = E_STR_UNKNOWN;
            fprintf(stderr, "%s: unknown string type '%s'\n",  prog_name, buf);
            continue;
        }
            
        /*
         * Put the ':' back and check the string's length.
         */
         
        buf[len] = ':';
        len      = strlen(buf);
        
        if (len < str->min_len)
        {
            ret_code = E_STR_SIZE;
            fprintf(stderr, "%s: string '%s' too short\n",  prog_name, buf);
        }
        else if (len > str->max_len)
        {
            ret_code = E_STR_SIZE;
            fprintf(stderr, "%s: string '%s' too long\n",  prog_name, buf);
        }
        
        /*
         * Tally how many of this string type we've seen.
         */
         
        str->count++; 
    }        

    /*
     * Check that the strings have occurred the correct number of times.
     */
     
    for (str = string_table; str->key != NULL; str++)
    {
        if (str->count < str->min_num)
        {
            ret_code = E_STR_COUNT;
            fprintf(stderr, "%s: too few type '%s' strings\n",  prog_name, str->key);
        }
        else if (str->count > str->max_num)
        {
            ret_code = E_STR_COUNT;
            fprintf(stderr, "%s: too many type '%s' strings\n",  prog_name, str->key);
        }
    }

   /*
    * Find the CRC.
    */
     
    temp_ret_code = find_crc(&file, &offset, &crc);
    
    /*
     * If the CRC was found Ok then calculate what the CRC should be and
     * compare.
     */
     
    if (temp_ret_code == E_OK)
    {
        if (generate_crc(&file, offset, CRC_MASK, CRC_MASK_SIZE) != crc)
        {
            temp_ret_code = E_INVALID_CRC;
        }
    }
    
    /*
     * Otherwise display an appropriate message for the return code.
     */
     
    if (temp_ret_code == E_NO_CRC)
    {
        ret_code = temp_ret_code;
        fprintf(stderr, "%s: the CRC was not found\n", prog_name);
    }
    
    else if (temp_ret_code == E_CRC_SIZE)
    {
        ret_code = temp_ret_code;
        fprintf(stderr, "%s: the CRC has an invalid format\n", prog_name);
    }
    
    else if (temp_ret_code == E_UNINIT_CRC)
    {
        ret_code = temp_ret_code;
        fprintf(stderr, "%s: the CRC has not been set\n", prog_name);
    }
    
    else if (temp_ret_code != E_OK)
    {
        ret_code = temp_ret_code;
        fprintf(stderr, "%s: invalid CRC - this file may be corrupt\n", prog_name);
    }
        
    /*
     * Unload the image.
     */
     
    free_file(&file);
    
    RETURN(ret_code);               
}

#endif


#if defined(BUILD_FULL)

/*----------------------------------------------------------------------------
|
| Function   - get_string
|
| Parameters - name     -> Name/path of the file to examine.
|              str      -> Pointer to the string table entry of the string
|                          to get.
|              buffer   -> Pointer to a buffer for the string.
|              buf_size -> Size of the buffer.
|
| Purpose    - Extract the first occurance of the specifed string from
|              the specifed file.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/

int
get_string(
    char       * name,
    STRING_OBJ * str,
    char       * buffer,
    size_t       buf_size
    )
{
    FILE_OBJ file;
    size_t   offset;
    char     prefix[MAX_STRING_SIZE];
    
    /*
     * Load the image file into core.
     */
     
    if (load_file(name, &file) != E_OK)
    {
        RETURNX();
    }
    
    /*
     * Find the first occurrance of the strings in the in-core image. If we
     * fail to find it then abort.
     */
     
    offset = 0;
        
    strcpy(prefix, mver_marker);
    strcat(prefix, str->key);
    strcat(prefix, ":");
        
    if (!find_string(&file, prefix, &offset))
    {
        free_file(&file);
        fprintf(stderr, "%s: string key '%s' not found\n", prog_name, str->key);
        RETURN(E_NO_STR);
    }
    
    /*
     * Extract the string from the image. This will set the error
     * code returned later.
     */
     
    extract_string(&file, offset, buffer, buf_size);
    
    /*
     * Unload the file.
     */
     
    free_file(&file);
    
    /*
     * Return the error code set by extract_string.
     */
     
    RETURNX();
}  

#endif


#if defined(BUILD_FULL) || defined(BUILD_RESTRICTED)

/*----------------------------------------------------------------------------
|
| Function   - set_string
|
| Parameters - name      -> Name/path of the file to patch.
|              str       -> Pointer to the string table entry of the string
|                           to set.
|              patch_str -> Pointer to a buffer containing the string.
|
| Purpose    - Patch the first occurance of the specified string in
|              the named file with the string in buffer. The CRC
|              is re-calculated.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/

int
set_string(
    char       * name,
    STRING_OBJ * str,
    char       * patch_str
    )
{
    FILE_OBJ file;
    size_t   offset;
    size_t   len;
    size_t   pos;
    char     prefix[MAX_STRING_SIZE];
    char     term;
    
    /*
     * Load the image file into core.
     */
     
    if (load_file(name, &file) != E_OK)
    {
        RETURNX();
    }
    
    /*
     * Find the first occurrance of the strings in the in-core image. If we
     * fail to find it then abort.
     */
     
    offset = 0;
        
    strcpy(prefix, mver_marker);
    strcat(prefix, str->key);
    strcat(prefix, ":");
        
    if (!find_string(&file, prefix, &offset))
    {
        free_file(&file);
        fprintf(stderr, "%s: string key '%s' not found\n", prog_name, str->key);
        RETURN(E_NO_STR);
    }

    /*
     * Check that the string format is valid by extracting it.
     */
     
    if (extract_string(&file, offset, prefix, MAX_STRING_SIZE) != E_OK)
    {
        free_file(&file);
        RETURNX();
    }
    
    /*
     * Work out what the terminator is (could be null or '@').
     */
     
    pos = offset;
    
    do
    {
        term = file.image[pos];
        pos++;
    }
    while (!IS_TERM(term));
    
    /*
     * Patch the new string in. We add a leading space and a terminator.
     * To be on the safe side we'll check that the file is long enough
     * for the new string.
     */
     
    len = strlen(patch_str);
     
    if (offset >= file.size - len - 2)
    {
        free_file(&file);
        fprintf(stderr, "%s: new string is too long for file\n", prog_name);
        RETURN(E_STR_SIZE);
    }
    
    file.image[offset]           = ' ';
    file.image[offset + len + 1] = term;
    memcpy(&file.image[offset + 1], patch_str, len);

    /*
     * Re-calculate the CRC.
     */
     
    set_crc(NULL, &file);
         
    /*
     * And save the file. save_file will set the error code returned.
     */
     
    save_file(&file);
    free_file(&file);
    
    RETURNX();
}

#endif


/*----------------------------------------------------------------------------
|
| Function   - show_usage
|
| Parameters - None.
|
| Purpose    - Display usage information.
|
| Returns    - Nothing.
|
|---------------------------------------------------------------------------*/

void
show_usage(
    void
    )
{
    fprintf(stderr, "\n");
    fprintf(stderr, "%s\n", mver_title + MVER_PREFIX_SIZE);
    fprintf(stderr, "%s\n", mver_copyright + MVER_PREFIX_SIZE);
    fprintf(stderr, "%s\n", mver_version + MVER_PREFIX_SIZE);
    
    if (mver_restricted != NULL)
    {
        fprintf(stderr, "%s\n", mver_restricted + MVER_PREFIX_SIZE);
    }
        
    fprintf(stderr, "\n");
    
    fprintf(stderr, "Usage: %s target\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "       Display product tracking strings.\n");
    fprintf(stderr, "\n");

#if defined(BUILD_FULL)

    fprintf(stderr, "       %s target -s\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "       Set the CRC in file 'target'.\n");
    fprintf(stderr, "\n");

    fprintf(stderr, "       %s target -c key source\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "       Set product tracking string with key 'key' in file\n");
    fprintf(stderr, "       'target' to be the same as that in file 'source'.\n");
    fprintf(stderr, "\n");
    
    fprintf(stderr, "       %s target -p key string\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "       Set product tracking string with key 'key' in file\n");
    fprintf(stderr, "       'target' to be the string 'string'.\n");
    fprintf(stderr, "\n");

#endif

#if defined(BUILD_FULL) || defined(BUILD_RESTRICTED)

    fprintf(stderr, "       %s target -r string\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "       Set product tracking string with key 'RESTRICTED' in\n");
    fprintf(stderr, "       file 'target' to be the string 'string'.\n");
    fprintf(stderr, "\n");

    fprintf(stderr, "       %s target -v\n", prog_name);
    fprintf(stderr, "\n");
    fprintf(stderr, "       Validate the product tracking strings in file 'target'.\n");
    fprintf(stderr, "\n");
    
#endif
}

        
/*----------------------------------------------------------------------------
|
| Function   - main
|
| Parameters - argc -> Number of command line arguments.
|              argv -> Pointer to a table of command line arguments.
|
| Purpose    - Main entry point.
|
| Returns    - An error code.
|
|---------------------------------------------------------------------------*/
            
int
main(
    int      argc,
    char * * argv
    )
{
    char         buffer[MAX_STRING_SIZE];
    int          i;
    char       * name;
    char       * source_name;
    char       * key_name;
    char       * patch_str;
    STRING_OBJ * str_ptr;
    int          mode;
    int          ret_code;

    /* 
     * Set up the magic marker strings. We build these strings
     * at run time to avoid confusing oursleves when applied
     * to ourselves.
     */
     
    mver_marker[0] = MVER_MARKER_FIRST;
    crc_marker[0]  = CRC_MARKER_FIRST;
         
    /*
     * Note the program name for error output. Some operating systems
     * put the path for argv[0], others just put the program name. We'll
     * skip the path if it's present.
     */
     
    for (i = strlen(argv[0]); i > 0; i--)
    {
        if (argv[0][i] == '.')
        {
            argv[0][i] = '\0';
        }
            
        if (argv[0][i] == ':' || argv[0][i] == '/' || argv[0][i] == '\\')
        {
            i++;
            break;
        }
    }
    
    prog_name = &argv[0][i];    
            
    /*
     * Parse the command line. We should have a file name followed by
     * a command and 0 to 2 arguments.
     */
     
    i           = 0;
    mode        = MODE_DISPLAY;
    name        = NULL;
    key_name    = NULL;
    source_name = NULL;
    patch_str   = NULL;
    str_ptr     = NULL;
    
    /*
     * Get the name of the target file.
     */
     
    if (++i >= argc)
    {
        show_usage();
        exit(E_BAD_ARGS);
    }
    
    name = argv[i];
    
    
#if defined(BUILD_FULL) || defined(BUILD_RESTRICTED)
        
    /*
     * Get the command. No command means just display the product
     * tracking strings.
     */
     
    if (++i < argc)
    {
        /*
         * Patch the restricted string mode. This is the same as patch string
         * mode but we hardwire the key. Should be one more arguments, 
         * a string to patch.
         */

        if (strcmp(argv[i], "-r") == 0)
        {
            key_name = RESTRICTED_KEY;
            
            if (++i < argc)
            {
                patch_str = argv[i];
            }
            
            mode = MODE_PATCH_RESTRICTED;
        }
    
        /*
         * Validate the strings. There shouldn't be any other arguments.
         */
         
        else if (strcmp(argv[i], "-v") == 0)
        {
            mode = MODE_VALIDATE;
        }

#if defined(BUILD_FULL)
        
        /*
         * Set the CRC mode. There shouldn't be any other arguments.
         */

        else if (strcmp(argv[i], "-s") == 0)
        {
            mode = MODE_SET_CRC;
        }
        
        /*
         * Copy a string mode. Should be two more arguments. A key name
         * and a source file name.
         */
         
        else if (strcmp(argv[i], "-c") == 0)
        {
            if (++i < argc)
            {
                key_name = argv[i];
            }
            
            if (++i < argc)
            {
                source_name = argv[i];
            }
            
            mode = MODE_COPY_STRING;
        }
        
        /*
         * Patch a string mode. Should be two more arguments. A key name
         * and a string to patch.
         */
        
        else if (strcmp(argv[i], "-p") == 0)
        {
            if (++i < argc)
            {
                key_name = argv[i];
            }
            
            if (++i < argc)
            {
                patch_str = argv[i];
            }
            
            mode = MODE_PATCH_STRING;
        }

#endif

        /*
         * A bad command.
         */
         
        else
        {
            show_usage();
            exit(E_BAD_ARGS);
        }                
    }
    
#endif
    
    /*
     * There shouldn't be any more arguments.
     */
         
    if (++i < argc)
    {
        show_usage();
        exit(E_BAD_ARGS);
    }                
    
    /*
     * We must have a target file name for all modes.
     */
     
    if (name == NULL)
    {
        show_usage();
        exit(E_BAD_ARGS);
    }

    /*
     * if we have been given a key name then validate it.
     */
     
    if (key_name != NULL)
    {
        str_ptr = lookup_key(key_name);
    }
             
    /*
     * Now process the target file according to the mode.     
     */

    /*
     * Display the product tracking strings.
     */
          
    if (mode == MODE_DISPLAY)
    {
        ret_code = display_strings(name);
    }

#if defined(BUILD_FULL)
    
    /*
     * Set the CRC.
     */
     
    else if (mode == MODE_SET_CRC)
    {
        ret_code = set_crc(name, NULL);
    }
    
    /*
     * Copy a product tracking string from one file to another.
     */
     
    else if (mode == MODE_COPY_STRING)
    {
        if (str_ptr == NULL || source_name == NULL)
        {
            show_usage();
            exit(E_BAD_ARGS);
        }
        
        if ((ret_code = get_string(
                            source_name, 
                            str_ptr, 
                            buffer, 
                            MAX_STRING_SIZE
                            )) == E_OK)
        {
            ret_code = set_string(name, str_ptr, buffer);
        }
    }

#endif

#if defined(BUILD_FULL) || defined(BUILD_RESTRICTED)
    
    /*
     * Patch a product tracking string.
     */
    
    else if (mode == MODE_PATCH_STRING || mode == MODE_PATCH_RESTRICTED)
    {
        if (str_ptr == NULL || patch_str == NULL)
        {
            show_usage();
            exit(E_BAD_ARGS);
        }
            
        /*
         * If the patch string is enclosed in quotes then remove
         * them.
         */
         
        if (patch_str[0] == '"' || patch_str[0] == '\'')
        {
            patch_str++;
        }
        
        if ((i = strlen(patch_str)) > 0)
        {
            if (patch_str[i] == '"' || patch_str[i] == '\'')
            {
                patch_str[i] = '\0';
            }
        }
                            
        ret_code = set_string(name, str_ptr, patch_str);
    }

    /*
     * Validate the product tracking strings.
     */

    else if (mode == MODE_VALIDATE)
    {
        ret_code = validate_strings(name);
    }

#endif
                    
    /*
     * And exit.
     */
     
    exit(ret_code);
}

     
/**** End of vertool.c ******************************************************/
